# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import itertools
import pytest

import MDAnalysis as mda


# grab all known writers
# sort so test order is predictable for parallel tests
writers = sorted(
    set(mda._MULTIFRAME_WRITERS.values())
    | set(mda._SINGLEFRAME_WRITERS.values()),
    key=lambda x: x.__name__,
)
known_ts_haters = [
    mda.coordinates.MOL2.MOL2Writer,
    mda.coordinates.PDB.PDBWriter,
    mda.coordinates.PDB.MultiPDBWriter,
    mda.coordinates.PQR.PQRWriter,
    mda.coordinates.PDBQT.PDBQTWriter,
    mda.coordinates.LAMMPS.DATAWriter,
    mda.coordinates.CRD.CRDWriter,
]


@pytest.mark.parametrize("writer", writers)
def test_ts_error(writer, tmpdir):
    u = mda.Universe.empty(10, trajectory=True)

    if writer == mda.coordinates.chemfiles.ChemfilesWriter:
        # chemfiles Writer exists but doesn't work without chemfiles
        if not mda.coordinates.chemfiles.check_chemfiles_version():
            pytest.skip("Chemfiles not available")
        fn = str(tmpdir.join("out.xtc"))
    elif writer == mda.coordinates.LAMMPS.DATAWriter:
        pytest.skip("DATAWriter requires integer atom types")
    elif (
        writer == mda.coordinates.H5MD.H5MDWriter
        and not mda.coordinates.H5MD.HAS_H5PY
    ):
        pytest.skip("skipping H5MDWriter test because h5py is not installed")
    else:
        fn = str(tmpdir.join("out.traj"))

    if (
        writer == mda.coordinates.PDB.PDBWriter
        or writer == mda.coordinates.PDB.MultiPDBWriter
    ):
        errmsg = "PDBWriter cannot write Timestep objects directly"
    else:
        errmsg = "neither an AtomGroup or Universe"

    with pytest.raises(TypeError, match=errmsg):
        with writer(fn, n_atoms=u.atoms.n_atoms) as w:
            w.write(u.trajectory.ts)


@pytest.mark.parametrize("writer", writers)
def test_write_with_atomgroup(writer, tmpdir):
    u = mda.Universe.empty(10, trajectory=True)

    if writer == mda.coordinates.chemfiles.ChemfilesWriter:
        # chemfiles Writer exists but doesn't work without chemfiles
        if not mda.coordinates.chemfiles.check_chemfiles_version():
            pytest.skip("Chemfiles not available")
        fn = str(tmpdir.join("out.xtc"))
    elif writer == mda.coordinates.MOL2.MOL2Writer:
        pytest.skip("MOL2 only writes MOL2 back out")
    elif writer == mda.coordinates.LAMMPS.DATAWriter:
        pytest.skip("DATAWriter requires integer atom types")
    elif (
        writer == mda.coordinates.H5MD.H5MDWriter
        and not mda.coordinates.H5MD.HAS_H5PY
    ):
        pytest.skip("skipping H5MDWriter test because h5py is not installed")
    else:
        fn = str(tmpdir.join("out.traj"))

    # H5MDWriter raises ValueError if the trajectory has no units and
    # convert_units is True
    convert_units = writer != mda.coordinates.H5MD.H5MDWriter

    with writer(fn, n_atoms=u.atoms.n_atoms, convert_units=convert_units) as w:
        w.write(u.atoms)


@pytest.mark.parametrize("writer", writers)
def test_write_with_universe(writer, tmpdir):
    u = mda.Universe.empty(10, trajectory=True)

    if writer == mda.coordinates.chemfiles.ChemfilesWriter:
        # chemfiles Writer exists but doesn't work without chemfiles
        if not mda.coordinates.chemfiles.check_chemfiles_version():
            pytest.skip("Chemfiles not available")
        fn = str(tmpdir.join("out.xtc"))
    elif writer == mda.coordinates.MOL2.MOL2Writer:
        pytest.skip("MOL2 only writes MOL2 back out")
    elif writer == mda.coordinates.LAMMPS.DATAWriter:
        pytest.skip("DATAWriter requires integer atom types")
    elif (
        writer == mda.coordinates.H5MD.H5MDWriter
        and not mda.coordinates.H5MD.HAS_H5PY
    ):
        pytest.skip("skipping H5MDWriter test because h5py is not installed")
    else:
        fn = str(tmpdir.join("out.traj"))

    # H5MDWriter raises ValueError if the trajectory has no units and
    # convert_units is True
    convert_units = writer != mda.coordinates.H5MD.H5MDWriter

    with writer(fn, n_atoms=u.atoms.n_atoms, convert_units=convert_units) as w:
        w.write(u.atoms)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""OpenMM structure I/O --- :mod:`MDAnalysis.converters.OpenMM`
================================================================


Read coordinates data from a
`OpenMM <http://docs.openmm.org/latest/api-python/generated/openmm.app.simulation.Simulation.html#openmm.app.simulation.Simulation>`_
:class:`openmm.app.simulation.Simulation` with :class:`OpenMMReader`
into a MDAnalysis Universe.

Also converts other objects within the
`OpenMM Application Layer <http://docs.openmm.org/latest/api-python/app.html>`_:

- `openmm.app.pdbfile.PDBFile <http://docs.openmm.org/latest/api-python/generated/openmm.app.pdbfile.PDBFile.html#openmm.app.pdbfile.PDBFile>`_
- `openmm.app.modeller.Modeller <http://docs.openmm.org/latest/api-python/generated/openmm.app.modeller.Modeller.html#openmm.app.modeller.Modeller>`_
- `openmm.app.pdbxfile.PDBxFile <http://docs.openmm.org/latest/api-python/generated/openmm.app.pdbxfile.PDBxFile.html#openmm.app.pdbxfile.PDBxFile>`_

Example
-------
OpenMM can read various file formats into OpenMM objects.
MDAnalysis can then convert some of these OpenMM objects into MDAnalysis Universe objects.

    >>> import openmm.app as app
    >>> import MDAnalysis as mda
    >>> from MDAnalysis.tests.datafiles import PDBX
    >>> pdbxfile = app.PDBxFile(PDBX)
    >>> mda.Universe(pdbxfile)
    <Universe with 60 atoms>



Classes
-------

.. autoclass:: OpenMMSimulationReader
   :members:

.. autoclass:: OpenMMAppReader
   :members:


"""


import numpy as np

from ..coordinates import base


class OpenMMSimulationReader(base.SingleFrameReaderBase):
    """Reader for OpenMM Simulation objects


    .. versionadded:: 2.0.0
    """

    format = "OPENMMSIMULATION"
    units = {
        "time": "ps",
        "length": "nm",
        "velocity": "nm/ps",
        "force": "kJ/(mol*nm)",
        "energy": "kJ/mol",
    }

    @staticmethod
    def _format_hint(thing):
        """Can this reader read *thing*?"""
        try:
            from openmm.app import Simulation
        except ImportError:
            try:  # pragma: no cover
                from simtk.openmm.app import Simulation
            except ImportError:
                return False
        else:
            return isinstance(thing, Simulation)

    def _read_first_frame(self):
        self.n_atoms = self.filename.topology.getNumAtoms()

        self.ts = self._mda_timestep_from_omm_context()

        if self.convert_units:
            self.convert_pos_from_native(self.ts._pos)
            self.ts.triclinic_dimensions = self.convert_pos_from_native(
                self.ts.triclinic_dimensions, inplace=False
            )
            self.ts.dimensions[3:] = _sanitize_box_angles(
                self.ts.dimensions[3:]
            )
            self.convert_velocities_from_native(self.ts._velocities)
            self.convert_forces_from_native(self.ts._forces)
            self.convert_time_from_native(self.ts.dt)

    def _mda_timestep_from_omm_context(self):
        """Construct Timestep object from OpenMM context"""
        try:
            import openmm.unit as u
        except ImportError:  # pragma: no cover
            import simtk.unit as u

        state = self.filename.context.getState(
            -1, getVelocities=True, getForces=True, getEnergy=True
        )

        n_atoms = self.filename.context.getSystem().getNumParticles()

        ts = self._Timestep(n_atoms, **self._ts_kwargs)
        ts.frame = 0
        ts.data["time"] = state.getTime()._value
        ts.data["potential_energy"] = (
            state.getPotentialEnergy().in_units_of(u.kilojoule / u.mole)._value
        )
        ts.data["kinetic_energy"] = (
            state.getKineticEnergy().in_units_of(u.kilojoule / u.mole)._value
        )
        ts.triclinic_dimensions = state.getPeriodicBoxVectors(
            asNumpy=True
        )._value
        ts.dimensions[3:] = _sanitize_box_angles(ts.dimensions[3:])
        ts.positions = state.getPositions(asNumpy=True)._value
        ts.velocities = state.getVelocities(asNumpy=True)._value
        ts.forces = state.getForces(asNumpy=True)._value

        return ts


class OpenMMAppReader(base.SingleFrameReaderBase):
    """Reader for OpenMM Application layer objects

    See also `the object definition in the OpenMM Application layer <http://docs.openmm.org/latest/api-python/generated/openmm.app.simulation.Simulation.html#openmm.app.simulation.Simulation>`_

    .. versionadded:: 2.0.0
    """

    format = "OPENMMAPP"
    units = {"time": "ps", "length": "nm"}

    @staticmethod
    def _format_hint(thing):
        """Can this reader read *thing*?"""
        try:
            from openmm import app
        except ImportError:
            try:  # pragma: no cover
                from simtk.openmm import app
            except ImportError:
                return False
        else:
            return isinstance(thing, (app.PDBFile, app.Modeller, app.PDBxFile))

    def _read_first_frame(self):
        self.n_atoms = self.filename.topology.getNumAtoms()

        self.ts = self._mda_timestep_from_omm_app()

        if self.convert_units:
            self.convert_pos_from_native(self.ts._pos)
            if self.ts.dimensions is not None:
                self.ts.triclinic_dimensions = self.convert_pos_from_native(
                    self.ts.triclinic_dimensions, inplace=False
                )
                self.ts.dimensions[3:] = _sanitize_box_angles(
                    self.ts.dimensions[3:]
                )

    def _mda_timestep_from_omm_app(self):
        """Construct Timestep object from OpenMM Application object"""

        omm_object = self.filename
        n_atoms = omm_object.topology.getNumAtoms()

        ts = self._Timestep(n_atoms, **self._ts_kwargs)
        ts.frame = 0
        if omm_object.topology.getPeriodicBoxVectors() is not None:
            ts.triclinic_dimensions = np.array(
                omm_object.topology.getPeriodicBoxVectors()._value
            )
            ts.dimensions[3:] = _sanitize_box_angles(ts.dimensions[3:])
        ts.positions = np.array(omm_object.getPositions()._value)

        return ts


def _sanitize_box_angles(angles):
    """Ensure box angles correspond to first quadrant

    See `discussion on unitcell angles <https://github.com/MDAnalysis/mdanalysis/pull/2917/files#r620558575>`_
    """
    inverted = 180 - angles

    return np.min(np.array([angles, inverted]), axis=0)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda
import numpy as np
import pytest
from MDAnalysis.coordinates.TNG import HAS_PYTNG
from MDAnalysis.lib.mdamath import triclinic_box
from numpy.testing import assert_allclose, assert_equal

from MDAnalysisTests.coordinates.base import (
    BaseReference,
    MultiframeReaderTest,
)
from MDAnalysisTests.datafiles import (
    COORDINATES_TNG,
    COORDINATES_TOPOLOGY,
    TNG_traj,
    TNG_traj_gro,
    TNG_traj_uneven_blocks,
    TNG_traj_vels_forces,
)


@pytest.mark.skipif(HAS_PYTNG, reason="pytng present")
def test_pytng_not_present_raises():
    with pytest.raises(ImportError, match="please install pytng"):
        _ = mda.Universe(TNG_traj_gro, TNG_traj)


@pytest.mark.skipif(HAS_PYTNG, reason="pytng present")
def test_parse_n_atoms_no_pytng():
    with pytest.raises(ImportError, match="please install pytng"):
        mda.coordinates.TNG.TNGReader.parse_n_atoms(TNG_traj)


@pytest.mark.skipif(not HAS_PYTNG, reason="pytng not installed")
class TNGReference(BaseReference):
    """Reference synthetic trajectory that was
    copied from test_xdr.TRReference"""

    def __init__(self):
        super(TNGReference, self).__init__()
        self.trajectory = COORDINATES_TNG
        self.topology = COORDINATES_TOPOLOGY
        self.reader = mda.coordinates.TNG.TNGReader
        self.writer = mda.coordinates.TNG.TNGReader.Writer
        self.ext = "tng"
        self.changing_dimensions = True
        self.prec = 4

        self.first_frame.velocities = self.first_frame.positions / 10
        self.first_frame.forces = self.first_frame.positions / 100

        self.second_frame.velocities = self.second_frame.positions / 10
        self.second_frame.forces = self.second_frame.positions / 100

        self.last_frame.velocities = self.last_frame.positions / 10
        self.last_frame.forces = self.last_frame.positions / 100

        self.jump_to_frame.velocities = self.jump_to_frame.positions / 10
        self.jump_to_frame.forces = self.jump_to_frame.positions / 100

    def iter_ts(self, i):
        ts = self.first_frame.copy()
        ts.positions = 2**i * self.first_frame.positions
        ts.velocities = ts.positions / 10
        ts.forces = ts.positions / 100
        ts.time = i
        ts.frame = i
        return ts


@pytest.mark.filterwarnings("ignore:Empty string")
@pytest.mark.filterwarnings("ignore:Stride of block")
@pytest.mark.skipif(not HAS_PYTNG, reason="pytng not installed")
class TestTNGCoordinatesTraj(MultiframeReaderTest):
    @staticmethod
    @pytest.fixture()
    def ref():
        return TNGReference()

    def test_get_writer_1(self, reader):
        with pytest.raises(
            NotImplementedError,
            match="There is currently no writer for TNG files",
        ):
            reader.Writer()

    def test_get_writer_2(self, reader):
        with pytest.raises(
            NotImplementedError,
            match="There is currently no writer for TNG files",
        ):
            reader.Writer()


@pytest.mark.skipif(not HAS_PYTNG, reason="pytng not installed")
def test_tng_traj_uneven_blocks():
    with pytest.raises(IOError, match="Strides of TNG special blocks"):
        _ = mda.Universe(TNG_traj_gro, TNG_traj_uneven_blocks)


@pytest.mark.filterwarnings("ignore:Failed read for block")
@pytest.mark.skipif(not HAS_PYTNG, reason="pytng not installed")
class TestTNGTraj(object):

    _n_atoms = 1000
    _n_frames = 101
    _stride = 5000

    prec = 5

    # these values all taken from GMX dump / 10 to get MDA units
    _pos_frame_0_first_3_atoms = (
        np.array(
            [
                [2.53300e00, 1.24400e00, 3.50600e00],
                [8.30000e-01, 2.54400e00, 3.44800e00],
                [1.09100e00, 1.10000e-01, 3.12900e00],
            ]
        )
        * 10
    )

    _pos_frame_100_first_3_atoms = (
        np.array(
            [
                [4.40000e-01, 3.89000e-01, 1.37400e00],
                [1.43200e00, 1.64900e00, 2.93900e00],
                [2.01500e00, 2.10300e00, 2.65700e00],
            ]
        )
        * 10
    )

    _box_frame_0 = (
        np.array(
            [
                3.60140e00,
                0.00000e00,
                0.00000e00,
                0.00000e00,
                3.60140e00,
                0.00000e00,
                0.00000e00,
                0.00000e00,
                3.60140e00,
            ]
        ).reshape(3, 3)
        * 10
    )

    _box_frame_100 = (
        np.array(
            [
                3.60140e00,
                0.00000e00,
                0.00000e00,
                0.00000e00,
                3.60140e00,
                0.00000e00,
                0.00000e00,
                0.00000e00,
                3.60140e00,
            ]
        ).reshape(3, 3)
        * 10
    )

    _box_frame_100 = (
        np.array(
            [
                3.58965e00,
                0.00000e00,
                0.00000e00,
                0.00000e00,
                3.58965e00,
                0.00000e00,
                0.00000e00,
                0.00000e00,
                3.58965e00,
            ]
        ).reshape(3, 3)
        * 10
    )

    @pytest.fixture(scope="class")
    def universe(self):
        return mda.Universe(TNG_traj_gro, TNG_traj)

    def test_n_atoms(self, universe):
        assert_equal(universe.trajectory.n_atoms, self._n_atoms)

    def test_n_frames(self, universe):
        assert_equal(
            universe.trajectory.n_frames,
            self._n_frames,
            "wrong number of frames in TNG file",
        )

    def test_block_names(self, universe):
        """Check the block names in the file"""
        assert "TNG_TRAJ_BOX_SHAPE" in universe.trajectory.blocks
        assert "TNG_TRAJ_POSITIONS" in universe.trajectory.blocks
        assert "TNG_GMX_LAMBDA" in universe.trajectory.blocks

    def test_special_blocks(self, universe):
        """Check the position and box special blocks are present"""
        assert "TNG_TRAJ_BOX_SHAPE" in universe.trajectory.special_blocks
        assert "TNG_TRAJ_POSITIONS" in universe.trajectory.special_blocks

    def test_additional_blocks(self, universe):
        """Check the lambda special block is present"""
        assert "TNG_GMX_LAMBDA" in universe.trajectory.additional_blocks

    def check_strides(self, universe):
        """Check the stride of trajectory frames in integrator steps"""
        assert universe.trajectory._global_stride == self._stride

    def test_initial_frame_is_0(self, universe):
        assert_equal(
            universe.trajectory.ts.frame,
            0,
            "initial frame is not 0 but {0}".format(
                universe.trajectory.ts.frame
            ),
        )

    def test_starts_with_first_frame(self, universe):
        """Test that coordinate arrays are filled as soon as the trajectory
        has been opened."""
        assert np.any(universe.atoms.positions > 0)

    def test_rewind(self, universe):
        trj = universe.trajectory
        trj.next()
        trj.next()  # for readers that do not support indexing
        assert_equal(
            trj.ts.frame, 2, "failed to forward to frame 2 (frameindex 2)"
        )
        trj.rewind()
        assert_equal(trj.ts.frame, 0, "failed to rewind to first frame")
        assert np.any(universe.atoms.positions > 0)

    def test_full_slice(self, universe):
        trj_iter = universe.trajectory[:]
        frames = [ts.frame for ts in trj_iter]
        assert_equal(frames, np.arange(universe.trajectory.n_frames))

    def test_random_access(self, universe):
        pos1 = universe.atoms[0].position
        universe.trajectory.next()
        universe.trajectory.next()
        pos3 = universe.atoms[0].position

        universe.trajectory[0]

        assert_equal(universe.atoms[0].position, pos1)

        universe.trajectory[2]

        assert_equal(universe.atoms[0].position, pos3)

    def test_frame_overrun(self, universe):
        with pytest.raises(IndexError, match="exceeds length of trajectory"):
            universe.trajectory[101]

    def test_positions_first_frame(self, universe):
        pos = universe.trajectory[0].positions
        assert_allclose(
            pos[0:3, :], self._pos_frame_0_first_3_atoms, rtol=10**-self.prec
        )

    def test_box_first_frame(self, universe):
        dims = universe.trajectory[0].dimensions
        assert_allclose(
            dims, triclinic_box(*self._box_frame_0), rtol=10**-self.prec
        )

    def test_positions_last_frame(self, universe):
        pos = universe.trajectory[100].positions
        assert_allclose(
            pos[0:3, :],
            self._pos_frame_100_first_3_atoms,
            rtol=10**-self.prec,
        )

    def test_box_last_frame(self, universe):
        dims = universe.trajectory[100].dimensions
        assert_allclose(
            dims, triclinic_box(*self._box_frame_100), rtol=10**-self.prec
        )

    @pytest.mark.parametrize("frame", [0, 20, 50, 100])
    def test_step(self, universe, frame):
        ts = universe.trajectory[frame]
        step = ts.data["step"]
        assert step == ts.frame * universe.trajectory._global_stride

    def test_lambda_in_ts(self, universe):
        ts = universe.trajectory[10]
        assert "TNG_GMX_LAMBDA" in ts.data.keys()
        assert isinstance(ts.data["TNG_GMX_LAMBDA"], np.ndarray)
        assert_equal(
            ts.data["TNG_GMX_LAMBDA"], np.asarray([[0]], dtype=np.float32)
        )

    def test_read_box_fail_strange_step(self, universe):
        stepnum = 123  # step number with no data
        iterator_step = universe.trajectory._file_iterator.read_step(stepnum)
        with pytest.raises(IOError, match="Failed to read box from TNG file"):
            universe.trajectory._frame_to_ts(
                iterator_step, universe.trajectory.ts
            )

    def test_read_pos_fail_strange_step(self, universe):
        stepnum = 123  # step number with no data
        iterator_step = universe.trajectory._file_iterator.read_step(stepnum)
        # set _has_box to False to trigger position reading error
        universe.trajectory._has_box = False
        with pytest.raises(
            IOError, match="Failed to read positions from TNG file"
        ):
            universe.trajectory._frame_to_ts(
                iterator_step, universe.trajectory.ts
            )

    def test_additional_block_read_fails(self, universe):
        stepnum = 123  # step number with no data
        iterator_step = universe.trajectory._file_iterator.read_step(stepnum)
        # set has_box, has_pos, to false to trigger GMX_LAMBDA reading error
        universe.trajectory._has_box = False
        universe.trajectory._has_positions = False
        # doesn't have velocities or forces
        with pytest.raises(
            IOError, match="Failed to read additional block TNG_GMX_LAMBDA"
        ):
            universe.trajectory._frame_to_ts(
                iterator_step, universe.trajectory.ts
            )

    def test_parse_n_atoms(self, universe):
        assert universe.trajectory.parse_n_atoms(TNG_traj) == self._n_atoms


@pytest.mark.filterwarnings("ignore:Off stride read for block")
@pytest.mark.skipif(not HAS_PYTNG, reason="pytng not installed")
class TestTNGTraj_vels_forces(object):

    _n_atoms = 1000
    _n_frames = 51
    _stride = 10

    prec = 5

    @pytest.fixture(scope="class")
    def universe(self):
        return mda.Universe(TNG_traj_gro, TNG_traj_vels_forces)

    def test_n_atoms(self, universe):
        assert_equal(universe.trajectory.n_atoms, self._n_atoms)

    def test_n_frames(self, universe):
        assert_equal(
            universe.trajectory.n_frames,
            self._n_frames,
            "wrong number of frames in TNG file",
        )

    def test_block_names(self, universe):
        """Check the block names in the file"""
        assert "TNG_TRAJ_BOX_SHAPE" in universe.trajectory.blocks
        assert "TNG_TRAJ_POSITIONS" in universe.trajectory.blocks
        assert "TNG_TRAJ_VELOCITIES" in universe.trajectory.blocks
        assert "TNG_TRAJ_FORCES" in universe.trajectory.blocks
        assert "TNG_GMX_LAMBDA" in universe.trajectory.blocks

    def test_special_blocks(self, universe):
        """Check the position and box special blocks are present"""
        assert "TNG_TRAJ_BOX_SHAPE" in universe.trajectory.special_blocks
        assert "TNG_TRAJ_POSITIONS" in universe.trajectory.special_blocks
        assert "TNG_TRAJ_VELOCITIES" in universe.trajectory.special_blocks
        assert "TNG_TRAJ_FORCES" in universe.trajectory.special_blocks

    def check_strides(self, universe):
        """Check the stride of trajectory frames in integrator steps"""
        assert universe.trajectory._global_stride == self._stride

    def test_read_vels_fail_strange_step(self, universe):
        stepnum = 123  # step number with no data
        iterator_step = universe.trajectory._file_iterator.read_step(stepnum)
        # set _has_* attrs to False to trigger velocities reading error
        universe.trajectory._has_box = False
        universe.trajectory._has_positions = False
        with pytest.raises(
            IOError, match="Failed to read velocities from TNG file"
        ):
            universe.trajectory._frame_to_ts(
                iterator_step, universe.trajectory.ts
            )

    def test_read_force_fail_strange_step(self, universe):
        stepnum = 123  # step number with no data
        iterator_step = universe.trajectory._file_iterator.read_step(stepnum)
        # set _has_* attrs to False to trigger forces reading error
        universe.trajectory._has_box = False
        universe.trajectory._has_positions = False
        universe.trajectory._has_velocities = False
        with pytest.raises(
            IOError, match="Failed to read forces from TNG file"
        ):
            universe.trajectory._frame_to_ts(
                iterator_step, universe.trajectory.ts
            )


@pytest.mark.skipif(not HAS_PYTNG, reason="pytng not installed")
def test_writer_raises_notimpl():
    u = mda.Universe(TNG_traj_gro, TNG_traj)
    with pytest.raises(
        NotImplementedError,
        match="There is currently no writer for TNG files",
    ):
        u.trajectory.Writer()



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import functools

import MDAnalysis as mda
import MDAnalysis.topology.TPRParser
import numpy as np
import pytest
# fmt: off
from MDAnalysis.tests.datafiles import (TPR, TPR400, TPR402, TPR403, TPR404,
                                        TPR405, TPR406, TPR407, TPR450, TPR451,
                                        TPR452, TPR453, TPR454, TPR455, TPR460,
                                        TPR461, TPR502, TPR504, TPR505, TPR510,
                                        TPR2016, TPR2018, TPR2019B3, TPR2020,
                                        TPR2020B2, TPR2021, TPR2022RC1,
                                        TPR2023, TPR2024, TPR2024_4,
                                        TPR2025_0,
                                        TPR_EXTRA_407, TPR_EXTRA_2016,
                                        TPR_EXTRA_2018, TPR_EXTRA_2020,
                                        TPR_EXTRA_2021, TPR_EXTRA_2022RC1,
                                        TPR_EXTRA_2023, TPR_EXTRA_2024,
                                        TPR_EXTRA_2024_4, XTC, TPR334_bonded,
                                        TPR_EXTRA_2025_0,
                                        TPR455Double, TPR510_bonded,
                                        TPR2016_bonded, TPR2018_bonded,
                                        TPR2019B3_bonded, TPR2020_bonded,
                                        TPR2020_double_bonded,
                                        TPR2020B2_bonded, TPR2020Double,
                                        TPR2021_bonded, TPR2021_double_bonded,
                                        TPR2021Double, TPR2022RC1_bonded,
                                        TPR2023_bonded, TPR2024_4_bonded,
                                        TPR2025_0_bonded, TPR2024_bonded,
                                        TPR_NNPOT_2025_0)
from numpy.testing import assert_equal

# fmt: on
from MDAnalysisTests.topology.base import ParserBase

BONDED_TPRS = (
    TPR510_bonded,
    TPR2016_bonded,
    TPR2018_bonded,
    TPR2019B3_bonded,
    TPR2021_bonded,
    TPR2021_double_bonded,
    TPR2020_bonded,
    TPR2020_double_bonded,
    TPR2022RC1_bonded,
    TPR2023_bonded,
    TPR2024_bonded,
    TPR2024_4_bonded,
    TPR2025_0_bonded,
    TPR_EXTRA_2025_0,
    TPR_EXTRA_2024_4,
    TPR_EXTRA_2024,
    TPR_EXTRA_2023,
    TPR_EXTRA_2022RC1,
    TPR_EXTRA_2021,
    TPR_EXTRA_2020,
    TPR_EXTRA_2018,
    TPR_EXTRA_2016,
    TPR_EXTRA_407,
)


class TPRAttrs(ParserBase):
    parser = MDAnalysis.topology.TPRParser.TPRParser
    expected_attrs = [
        "ids",
        "types",
        "masses",
        "names",
        "elements",
        "resids",
        "resnames",
        "chainIDs",
        "moltypes",
        "molnums",
        "charges",
        "bonds",
        "angles",
        "dihedrals",
        "impropers",
    ]

    def test_moltypes(self, top):
        moltypes = top.moltypes.values
        assert_equal(moltypes, self.ref_moltypes)

    def test_molnums(self, top):
        molnums = top.molnums.values
        assert_equal(molnums, self.ref_molnums)
        assert molnums.dtype == np.intp

    def test_chainIDs(self, top):
        if hasattr(self, "ref_chainIDs"):
            assert_equal(
                self.ref_chainIDs, getattr(top, "chainIDs").name_lookup
            )


class TestTPR(TPRAttrs):
    """
    this test the data/adk_oplsaa.tpr which is of tpx version 58
    """

    expected_n_atoms = 47681
    expected_n_residues = 11302
    expected_n_segments = 3
    ref_moltypes = np.array(
        ["AKeco"] * 214 + ["SOL"] * 11084 + ["NA+"] * 4, dtype=object
    )
    ref_molnums = np.array([0] * 214 + list(range(1, 1 + 11084 + 4)))

    @pytest.fixture()
    def filename(self):
        return TPR


# The follow test the same system grompped by different version of gromacs
# FORMAT: TPRABC, where numbers ABC indicates the version of gromacs that
# generates the corresponding tpr file
class TestTPRGromacsVersions(TPRAttrs):
    expected_n_atoms = 2263
    expected_n_residues = 230
    expected_n_segments = 2
    ref_moltypes = np.array(["Protein_A"] * 129 + ["SOL"] * 101, dtype=object)
    ref_molnums = np.array([0] * 129 + list(range(1, 1 + 101)))
    ref_chainIDs = ["Protein_A", "SOL"]

    # fmt: off
    @pytest.fixture(params=[
            TPR400, TPR402, TPR403, TPR404, TPR405, TPR406, TPR407, TPR450,
            TPR451, TPR452, TPR453, TPR454, TPR455, TPR502, TPR504, TPR505,
            TPR510, TPR2016, TPR2018, TPR2019B3, TPR2020, TPR2020Double,
            TPR2021, TPR2021Double, TPR2022RC1, TPR2023, TPR2024, TPR2024_4,
            TPR2025_0,
        ]
    )
    # fmt: on
    def filename(self, request):
        return request.param


class TestTPRDouble(TPRAttrs):
    expected_n_atoms = 21692
    expected_n_residues = 4352
    expected_n_segments = 7
    ref_moltypes = np.array(
        ["DOPC"] * 21
        + ["DPPC"] * 10
        + ["CHOL"] * 3
        + ["DOPC"] * 21
        + ["DPPC"] * 10
        + ["CHOL"] * 3
        + ["SOL"] * 4284,
        dtype=object,
    )
    ref_molnums = np.arange(4352)

    @pytest.fixture()
    def filename(self):
        return TPR455Double


class TestTPR46x(TPRAttrs):
    expected_n_atoms = 44052
    expected_n_residues = 10712
    expected_n_segments = 8
    ref_moltypes = np.array(
        ["Protein_A"] * 27
        + ["Protein_B"] * 27
        + ["Protein_C"] * 27
        + ["Protein_D"] * 27
        + ["Protein_E"] * 27
        + ["SOL"] * 10530
        + ["NA+"] * 26
        + ["CL-"] * 21,
        dtype=object,
    )
    ref_molnums = np.array(
        [0] * 27
        + [1] * 27
        + [2] * 27
        + [3] * 27
        + [4] * 27
        + list(range(5, 5 + 10530 + 26 + 21))
    )

    @pytest.fixture(params=[TPR460, TPR461])
    def filename(self, request):
        return request.param


class TestTPRNnpot(TPRAttrs):
    expected_n_atoms = 23
    expected_n_residues = 2
    expected_n_segments = 1
    ref_moltypes = np.array(["Protein_chain_A"] * 2, dtype=object)
    ref_molnums = np.array([0] * 2)
    ref_chainIDs = ["A"]

    @pytest.fixture(params=[TPR_NNPOT_2025_0])
    def filename(self, request):
        return request.param


def _test_is_in_topology(name, elements, topology_path, topology_section):
    """
    Test if an interaction appears as expected in the topology
    """
    post_40_potentials = {
        "RESTRAINTPOT",
        "RESTRANGLES",
        "RESTRDIHS",
        "CBTDIHS",
        "PIDIHS",
    }
    if name in post_40_potentials and topology_path == TPR_EXTRA_407:
        # The potential is not yet implemented in this version of gromacs
        return
    parser = MDAnalysis.topology.TPRParser.TPRParser(topology_path)
    top = parser.parse()
    for element in elements:
        assert (
            element in getattr(top, topology_section).values
        ), 'Interaction type "{}" not found'.format(name)


@pytest.mark.parametrize("topology", BONDED_TPRS)
@pytest.mark.parametrize(
    "bond",
    (
        ("BONDS", [(0, 1)]),
        ("G96BONDS", [(1, 2)]),
        ("MORSE", [(2, 3)]),
        ("CUBICBONDS", [(3, 4)]),
        ("CONNBONDS", [(4, 5)]),
        ("HARMONIC", [(5, 6)]),
        ("FENEBONDS", [(6, 7)]),
        ("RESTRAINTPOT", [(7, 8)]),
        ("TABBONDS", [(8, 9)]),
        ("TABBONDSNC", [(9, 10)]),
        ("CONSTR", [(10, 11)]),
        ("CONSTRNC", [(11, 12)]),
    ),
)
def test_all_bonds(topology, bond):
    """Test that all bond types are parsed as expected"""
    bond_type_in_topology = functools.partial(
        _test_is_in_topology, topology_section="bonds"
    )
    bond_type, elements = bond
    bond_type_in_topology(bond_type, elements, topology)


@pytest.mark.parametrize("topology", BONDED_TPRS)
@pytest.mark.parametrize(
    "angle",
    (
        ("ANGLES", [(0, 1, 2)]),
        ("G96ANGLES", [(1, 2, 3)]),
        ("CROSS_BOND_BOND", [(2, 3, 4)]),
        ("CROSS_BOND_ANGLE", [(3, 4, 5)]),
        ("UREY_BRADLEY", [(4, 5, 6)]),
        ("QANGLES", [(5, 6, 7)]),
        ("RESTRANGLES", [(6, 7, 8)]),
        ("TABANGLES", [(7, 8, 9)]),
    ),
)
def test_all_angles(topology, angle):
    angle_type_in_topology = functools.partial(
        _test_is_in_topology, topology_section="angles"
    )
    angle_type, elements = angle
    angle_type_in_topology(angle_type, elements, topology)


@pytest.mark.parametrize("topology", BONDED_TPRS)
@pytest.mark.parametrize(
    "dih",
    (
        ("PDIHS", [(0, 1, 2, 3), (1, 2, 3, 4), (7, 8, 9, 10)]),
        ("RBDIHS", [(4, 5, 6, 7)]),
        ("RESTRDIHS", [(8, 9, 10, 11)]),
        ("CBTDIHS", [(9, 10, 11, 12)]),
        ("FOURDIHS", [(6, 7, 8, 9)]),
        ("TABDIHS", [(10, 11, 12, 13)]),
    ),
)
def test_all_dihedrals(topology, dih):
    dih_type_in_topology = functools.partial(
        _test_is_in_topology, topology_section="dihedrals"
    )
    dih_type, elements = dih
    dih_type_in_topology(dih_type, elements, topology)


@pytest.mark.parametrize("topology", BONDED_TPRS)
@pytest.mark.parametrize(
    "impr",
    (("IDIHS", [(2, 3, 4, 5), (3, 4, 5, 6)]), ("PIDIHS", [(5, 6, 7, 8)])),
)
def test_all_impropers(topology, impr):
    impr_type_in_topology = functools.partial(
        _test_is_in_topology, topology_section="impropers"
    )

    impr_type, elements = impr
    impr_type_in_topology(impr_type, elements, topology)


# fmt: off
@pytest.fixture(params=(
        TPR400, TPR402, TPR403, TPR404, TPR405, TPR406, TPR407, TPR450,
        TPR451, TPR452, TPR453, TPR454, TPR502, TPR504, TPR505, TPR510,
        TPR2016, TPR2018, TPR2023, TPR2024, TPR2024_4, TPR2025_0,
    )
)
# fmt: on
def bonds_water(request):
    parser = MDAnalysis.topology.TPRParser.TPRParser(request.param).parse()
    # The index of the first water atom is 1960
    first = 1960
    bonds = [
        bond
        for bond in parser.bonds.values
        if bond[0] >= first and bond[1] >= first
    ]
    return bonds


def test_settle(bonds_water):
    # There are 101 water molecule with 2 bonds each
    assert len(bonds_water) == 202
    # The last index corresponds to the last water atom
    assert bonds_water[-1][1] == 2262


@pytest.mark.parametrize("tpr_path", [TPR_NNPOT_2025_0])
def test_ala2(tpr_path):
    topology = MDAnalysis.topology.TPRParser.TPRParser(tpr_path).parse()
    # Check that bonds etc are read correctly
    assert len(topology.bonds.values) == 22  # 12 Bond + 10 Connect
    assert len(topology.angles.values) == 21  # 21 Angle
    assert len(topology.dihedrals.values) == 26  # 26 Proper Dih.
    assert len(topology.impropers.values) == 1  # 1 Per. Imp. Dih.


@pytest.mark.parametrize(
    "tpr_path, expected_exception",
    (
        (TPR2020B2, IOError),  # Gromacs 2020 beta see issue #2428
        (TPR2020B2_bonded, IOError),  # Gromacs 2020 beta see issue #2428
        (TPR334_bonded, NotImplementedError),  # Too old
        (XTC, IOError),  # Not a TPR file
    ),
)
def test_fail_for_unsupported_files(tpr_path, expected_exception):
    parser = MDAnalysis.topology.TPRParser.TPRParser(tpr_path)
    with pytest.raises(expected_exception):
        parser.parse()


@pytest.mark.parametrize("tpr_path", BONDED_TPRS)
def test_no_elements(tpr_path):
    """
    If the TPR does not contain element information, the element topology
    attribute is not defined.
    """
    parser = MDAnalysis.topology.TPRParser.TPRParser(tpr_path)
    topology = parser.parse()
    with pytest.raises(AttributeError):
        _ = topology.elements


def test_elements():
    tpr_path = TPR
    parser = MDAnalysis.topology.TPRParser.TPRParser(tpr_path)
    topology = parser.parse()
    reference = np.array(
        (
            "H,C,H,H,C,H,H,H,C,H,H,H,C,O,N,H,C,H,C,H,H,C,H,C,H,H,H,C,H,H,"
            "H,C,O,N,H,C,H,H,C,O,O,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H"
            ",H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,"
            "O,H,H"
        ).split(","),
        dtype=object,
    )
    assert_equal(topology.elements.values[3300:3400], reference)
    # fmt: off
    reference = np.array(
        [
            "O", "H", "H", "", "O", "H", "H", "", "O", "H", "H", "", "O", "H",
            "H", "", "Na", "Na", "Na", "Na",
        ],
        dtype=object,
    )
    # fmt: on
    assert_equal(topology.elements.values[-20:], reference)


@pytest.mark.parametrize(
    "resid_from_one,resid_addition",
    [
        (False, 0),
        (True, 1),  # status quo for 2.x
    ],
)
def test_resids(resid_from_one, resid_addition):
    u = mda.Universe(TPR, tpr_resid_from_one=resid_from_one)
    resids = np.arange(len(u.residues)) + resid_addition
    assert_equal(
        u.residues.resids,
        resids,
        err_msg="tpr_resid_from_one kwarg not switching resids",
    )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""Hydrogen Bond Analysis --- :mod:`MDAnalysis.analysis.hydrogenbonds.hbond_analysis`
=====================================================================================

:Author: Paul Smith
:Year: 2019
:Copyright: Lesser GNU Public License v2.1+

.. versionadded:: 1.0.0

This module provides methods to find and analyse hydrogen bonds in a Universe.

The :class:`HydrogenBondAnalysis` class is a new version of the original
:class:`MDAnalysis.analysis.hbonds.HydrogenBondAnalysis` class from the module
:mod:`MDAnalysis.analysis.hbonds.hbond_analysis`, which itself was modeled after the `VMD
HBONDS plugin`_.

.. _`VMD HBONDS plugin`: http://www.ks.uiuc.edu/Research/vmd/plugins/hbonds/


Input
------

Required:
  - *universe* : an MDAnalysis Universe object

Options:
  - *donors_sel* [None] : Atom selection for donors. If `None`, then will be identified via the topology.
  - *hydrogens_sel* [None] : Atom selection for hydrogens. If `None`, then will be identified via charge and mass.
  - *acceptors_sel* [None] : Atom selection for acceptors. If `None`, then will be identified via charge.
  - *d_h_cutoff* (Å) [1.2] : Distance cutoff used for finding donor-hydrogen pairs
  - *d_a_cutoff* (Å) [3.0] : Distance cutoff for hydrogen bonds. This cutoff refers to the D-A distance.
  - *d_h_a_angle_cutoff* (degrees) [150] : D-H-A angle cutoff for hydrogen bonds.
  - *update_selections* [True] : If true, will update atom selections at each frame.


Output
------

- *frame* : frame at which a hydrogen bond was found
- *donor id* : atom id of the hydrogen bond donor atom
- *hydrogen id* : atom id of the hydrogen bond hydrogen atom
- *acceptor id* : atom id of the hydrogen bond acceptor atom
- *distance* (Å): length of the hydrogen bond
- *angle* (degrees): angle of the hydrogen bond

Hydrogen bond data are returned in a :class:`numpy.ndarray` on a "one line, one observation" basis
and can be accessed via :attr:`HydrogenBondAnalysis.results.hbonds`::

    results = [
        [
            <frame>,
            <donor index (0-based)>,
            <hydrogen index (0-based)>,
            <acceptor index (0-based)>,
            <distance>,
            <angle>
        ],
        ...
    ]

Example use of :class:`HydrogenBondAnalysis`
--------------------------------------------

The simplest use case is to allow :class:`HydrogenBondAnalysis` to guess the acceptor and hydrogen atoms, and to
identify donor-hydrogen pairs via the bonding information in the topology::

  import MDAnalysis
  from MDAnalysis.analysis.hydrogenbonds.hbond_analysis import HydrogenBondAnalysis as HBA

  u = MDAnalysis.Universe(psf, trajectory)

  hbonds = HBA(universe=u)
  hbonds.run()

It is also possible to specify which hydrogens and acceptors to use in the analysis. For example, to find all hydrogen
bonds in water::

  import MDAnalysis
  from MDAnalysis.analysis.hydrogenbonds.hbond_analysis import HydrogenBondAnalysis as HBA

  u = MDAnalysis.Universe(psf, trajectory)

  hbonds = HBA(universe=u, hydrogens_sel='resname TIP3 and name H1 H2', acceptors_sel='resname TIP3 and name OH2')
  hbonds.run()

Alternatively, :attr:`hydrogens_sel` and :attr:`acceptors_sel` may be generated via the :attr:`guess_hydrogens` and
:attr:`guess_acceptors`. This selection strings may then be modified prior to calling :attr:`run`, or a subset of
the universe may be used to guess the atoms. For example, find hydrogens and acceptors belonging to a protein::

  import MDAnalysis
  from MDAnalysis.analysis.hydrogenbonds.hbond_analysis import HydrogenBondAnalysis as HBA

  u = MDAnalysis.Universe(psf, trajectory)

  hbonds = HBA(universe=u)
  hbonds.hydrogens_sel = hbonds.guess_hydrogens("protein")
  hbonds.acceptors_sel = hbonds.guess_acceptors("protein")
  hbonds.run()

Slightly more complex selection strings are also possible. For example, to find hydrogen bonds involving a protein and
any water molecules within 10 Å of the protein (which may be useful for subsequently finding the lifetime of
protein-water hydrogen bonds or finding water-bridging hydrogen bond paths)::

  import MDAnalysis
  from MDAnalysis.analysis.hydrogenbonds.hbond_analysis import HydrogenBondAnalysis as HBA

  u = MDAnalysis.Universe(psf, trajectory)

  hbonds = HBA(universe=u)

  protein_hydrogens_sel = hbonds.guess_hydrogens("protein")
  protein_acceptors_sel = hbonds.guess_acceptors("protein")

  water_hydrogens_sel = "resname TIP3 and name H1 H2"
  water_acceptors_sel = "resname TIP3 and name OH2"

  hbonds.hydrogens_sel = f"({protein_hydrogens_sel}) or ({water_hydrogens_sel} and around 10 not resname TIP3})"
  hbonds.acceptors_sel = f"({protein_acceptors_sel}) or ({water_acceptors_sel} and around 10 not resname TIP3})"
  hbonds.run()

To calculate the hydrogen bonds between different groups, for example a
protein and water, one can use the :attr:`between` keyword. The
following will find protein-water hydrogen bonds but not protein-protein
or water-water hydrogen bonds::

  import MDAnalysis
  from MDAnalysis.analysis.hydrogenbonds.hbond_analysis import (
    HydrogenBondAnalysis as HBA)

  u = MDAnalysis.Universe(psf, trajectory)

  hbonds = HBA(
    universe=u,
    between=['resname TIP3', 'protein']
    )

  protein_hydrogens_sel = hbonds.guess_hydrogens("protein")
  protein_acceptors_sel = hbonds.guess_acceptors("protein")

  water_hydrogens_sel = "resname TIP3 and name H1 H2"
  water_acceptors_sel = "resname TIP3 and name OH2"

  hbonds.hydrogens_sel = f"({protein_hydrogens_sel}) or ({water_hydrogens_sel})"
  hbonds.acceptors_sel = f"({protein_acceptors_sel}) or ({water_acceptors_sel})"

  hbonds.run()

It is further possible to compute hydrogen bonds between several groups with
with use of :attr:`between`. If in the above example,
`between=[['resname TIP3', 'protein'], ['protein', 'protein']]`, all
protein-water and protein-protein hydrogen bonds will be found, but
no water-water hydrogen bonds.

One can also define hydrogen bonds with atom types::

  from MDAnalysis.analysis.hydrogenbonds.hbond_analysis import HydrogenBondAnalysis as HBA
  hbonds = HBA(
               universe=u,
               donors_sel='type 2',
               hydrogens_sel='type 1',
               acceptors_sel='type 2',
              )

In order to compute the hydrogen bond lifetime, after finding hydrogen bonds
one can use the :attr:`lifetime` function::

    ...
    hbonds.run()
    tau_timeseries, timeseries = hbonds.lifetime()

It is **highly recommended** that a topology with bond information is used to
generate the universe, e.g `PSF`, `TPR`, or `PRMTOP` files. This is the only
method by which it can be guaranteed that donor-hydrogen pairs are correctly
identified. However, if, for example, a `PDB` file is used instead, a
:attr:`donors_sel` may be provided along with a :attr:`hydrogens_sel` and the
donor-hydrogen pairs will be identified via a distance cutoff,
:attr:`d_h_cutoff`::

  import MDAnalysis
  from MDAnalysis.analysis.hydrogenbonds.hbond_analysis import (
    HydrogenBondAnalysis as HBA)

  u = MDAnalysis.Universe(pdb, trajectory)

  hbonds = HBA(
    universe=u,
    donors_sel='resname TIP3 and name OH2',
    hydrogens_sel='resname TIP3 and name H1 H2',
    acceptors_sel='resname TIP3 and name OH2',
    d_h_cutoff=1.2
  )
  hbonds.run()

The class and its methods
-------------------------

.. autoclass:: HydrogenBondAnalysis
   :members:
   :inherited-members: run

   .. attribute:: results.hbonds

      A :class:`numpy.ndarray` which contains a list of all observed hydrogen
      bond interactions. See `Output`_ for more information.

      .. versionadded:: 2.0.0

   .. attribute:: hbonds

      Alias to the :attr:`results.hbonds` attribute.

      .. deprecated:: 2.0.0
         Will be removed in MDAnalysis 3.0.0. Please use
         :attr:`results.hbonds` instead.
"""
import logging
import warnings
from collections.abc import Iterable

import numpy as np

from ..base import AnalysisBase, Results, ResultsGroup
from MDAnalysis.lib.distances import capped_distance, calc_angles
from MDAnalysis.lib.correlations import autocorrelation, correct_intermittency
from MDAnalysis.exceptions import NoDataError
from MDAnalysis.core.groups import AtomGroup
from MDAnalysis.analysis.hydrogenbonds.hbond_autocorrel import find_hydrogen_donors

from ...due import due, Doi


logger = logging.getLogger(__name__)


due.cite(Doi("10.1039/C9CP01532A"),
         description="Hydrogen bond analysis implementation",
         path="MDAnalysis.analysis.hydrogenbonds.hbond_analysis",
         cite_module=True)

del Doi


class HydrogenBondAnalysis(AnalysisBase):
    """
    Perform an analysis of hydrogen bonds in a Universe.
    """

    _analysis_algorithm_is_parallelizable = True

    @classmethod
    def get_supported_backends(cls):
        return ('serial', 'multiprocessing', 'dask',)

    def __init__(self, universe,
                 donors_sel=None, hydrogens_sel=None, acceptors_sel=None,
                 between=None, d_h_cutoff=1.2,
                 d_a_cutoff=3.0, d_h_a_angle_cutoff=150,
                 update_selections=True):
        """Set up atom selections and geometric criteria for finding hydrogen
        bonds in a Universe.

        Hydrogen bond selections with `donors_sel` , `hydrogens_sel`, and
        `acceptors_sel` may be achieved with either a *resname*, atom *name* 
        combination, or when those are absent, with atom *type* selections.

        Parameters
        ----------
        universe : Universe
            MDAnalysis Universe object
        donors_sel : str
            Selection string for the hydrogen bond donor atoms. If the
            universe topology contains bonding information, leave
            :attr:`donors_sel` as `None` so that donor-hydrogen pairs can be
            correctly identified.
        hydrogens_sel :  str
            Selection string for the hydrogen bond hydrogen atoms. Leave as
            `None` to guess which hydrogens to use in the analysis using
            :attr:`guess_hydrogens`. If :attr:`hydrogens_sel` is left as
            `None`, also leave :attr:`donors_sel` as None so that
            donor-hydrogen pairs can be correctly identified.
        acceptors_sel : str
            Selection string for the hydrogen bond acceptor atoms. Leave as
            `None` to guess which atoms to use in the analysis using
            :attr:`guess_acceptors`
        between : List (optional),
            Specify two selection strings for non-updating atom groups between
            which hydrogen bonds will be calculated. For example, if the donor
            and acceptor selections include both protein and water, it is
            possible to find only protein-water hydrogen bonds - and not
            protein-protein or water-water - by specifying
            between=["protein", "SOL"]`. If a two-dimensional list is
            passed, hydrogen bonds between each pair will be found. For
            example, between=[["protein", "SOL"], ["protein", "protein"]]`
            will calculate all protein-water and protein-protein hydrogen
            bonds but not water-water hydrogen bonds. If `None`, hydrogen
            bonds between all donors and acceptors will be calculated.
        d_h_cutoff : float (optional)
            Distance cutoff used for finding donor-hydrogen pairs.
            Only used to find donor-hydrogen pairs if the
            universe topology does not contain bonding information
        d_a_cutoff : float (optional)
            Distance cutoff for hydrogen bonds. This cutoff refers to the D-A distance.
        d_h_a_angle_cutoff : float (optional)
            D-H-A angle cutoff for hydrogen bonds, in degrees.
        update_selections : bool (optional)
            Whether or not to update the acceptor, donor and hydrogen
            lists at each frame.

        Note
        ----

        It is highly recommended that a universe topology with bond
        information is used, as this is the only way that guarantees the
        correct identification of donor-hydrogen pairs.


        .. versionadded:: 2.0.0
            Added `between` keyword
        .. versionchanged:: 2.4.0
            Added use of atom types in selection strings for hydrogen atoms, 
            bond donors, or bond acceptors
        .. versionchanged:: 2.8.0
            Introduced :meth:`get_supported_backends` allowing for parallel execution on
            :mod:`multiprocessing` and :mod:`dask` backends.
        .. versionchanged:: 2.10.0
            The `donors_sel`, `hydrogens_sel`, and `acceptors_sel` are stored as properties
            now and update the internal `_donor`, `_hydrogens`, and `_acceptors` when they
            are modified.
        """

        self.u = universe
        self._trajectory = self.u.trajectory

        self._donors_sel = donors_sel.strip() if donors_sel is not None else donors_sel
        self._hydrogens_sel = hydrogens_sel.strip() if hydrogens_sel is not None else hydrogens_sel
        self._acceptors_sel = acceptors_sel.strip() if acceptors_sel is not None else acceptors_sel

        msg = ("{} is an empty selection string - no hydrogen bonds will "
               "be found. This may be intended, but please check your "
               "selection."
               )
        for sel in ['donors_sel', 'hydrogens_sel', 'acceptors_sel']:
            val = getattr(self, sel)
            if isinstance(val, str) and not val:
                warnings.warn(msg.format(sel))

        # If hydrogen bonding groups are selected, then generate
        # corresponding atom groups
        if between is not None:
            if not isinstance(between, Iterable) or len(between) == 0:
                raise ValueError("between must be a non-empty list/iterable")
            if isinstance(between[0], str):
                between = [between]

            between_ags = []
            for group1, group2 in between:
                between_ags.append(
                    [
                        self.u.select_atoms(group1, updating=False),
                        self.u.select_atoms(group2, updating=False)
                    ]
                )

            self.between_ags = between_ags
        else:
            self.between_ags = None


        self.d_h_cutoff = d_h_cutoff
        self.d_a_cutoff = d_a_cutoff
        self.d_h_a_angle = d_h_a_angle_cutoff
        self.update_selections = update_selections
        self.results = Results()
        self.results.hbonds = None

        # Set atom selections if they have not been provided
        if self._acceptors_sel is None:
            self._acceptors_sel = self.guess_acceptors()
        if self._hydrogens_sel is None:
            self._hydrogens_sel = self.guess_hydrogens()

        # Select atom groups
        self._acceptors = self.u.select_atoms(self.acceptors_sel,
                                              updating=self.update_selections)
        self._donors, self._hydrogens = self._get_dh_pairs()

    def guess_hydrogens(self,
                        select='all',
                        max_mass=1.1,
                        min_charge=0.3,
                        min_mass=0.9
                        ):
        """Guesses which hydrogen atoms should be used in the analysis.

        Parameters
        ----------
        select: str (optional)
            :ref:`Selection string <selection-commands-label>` for atom group 
            from which hydrogens will be identified. (e.g., ``(resname X and 
            name H1)`` or ``type 2``)
        max_mass: float (optional)
            The mass of a hydrogen atom must be less than this value.
        min_mass: float (optional)
            The mass of a hydrogen atom must be greater than this value.
        min_charge: float (optional)
            The charge of a hydrogen atom must be greater than this value.

        Returns
        -------
        potential_hydrogens: str
            String containing the :attr:`resname` and :attr:`name` of all 
            hydrogen atoms potentially capable of forming hydrogen bonds.

        Notes
        -----
        Hydrogen selections may be achieved with either a resname, atom 
        name combination, or when those are absent, atom types.

        This function makes use of atomic masses and atomic charges to identify
        which atoms are hydrogen atoms that are capable of participating in 
        hydrogen bonding. If an atom has a mass less than :attr:`max_mass` and 
        an atomic charge greater than :attr:`min_charge` then it is considered 
        capable of participating in hydrogen bonds.

        If :attr:`hydrogens_sel` is `None`, this function is called to guess 
        the selection.

        Alternatively, this function may be used to quickly generate a 
        :class:`str` of potential hydrogen atoms involved in hydrogen bonding.
        This str may then be modified before being used to set the attribute
        :attr:`hydrogens_sel`.


        .. versionchanged:: 2.4.0
            Added ability to use atom types

        """

        if min_mass > max_mass:
            raise ValueError("min_mass is higher than (or equal to) max_mass")

        ag = self.u.select_atoms(select)
        hydrogens_ag = ag[
            np.logical_and.reduce((
                ag.masses < max_mass,
                ag.charges > min_charge,
                ag.masses > min_mass,
            ))
        ]

        return self._group_categories(hydrogens_ag)

    def guess_donors(self, select='all', max_charge=-0.5):
        """Guesses which atoms could be considered donors in the analysis. Only
        use if the universe topology does not contain bonding information, 
        otherwise donor-hydrogen pairs may be incorrectly assigned.

        Parameters
        ----------
        select: str (optional)
            :ref:`Selection string <selection-commands-label>` for atom group 
            from which donors will be identified. (e.g., ``(resname X and name 
            O1)`` or ``type 2``)
        max_charge: float (optional)
            The charge of a donor atom must be less than this value.

        Returns
        -------
        potential_donors: str
            String containing the :attr:`resname` and :attr:`name` of all atoms 
            that are potentially capable of forming hydrogen bonds.

        Notes
        -----
        Donor selections may be achieved with either a resname, atom 
        name combination, or when those are absent, atom types.

        This function makes use of and atomic charges to identify which atoms 
        could be considered donor atoms in the hydrogen bond analysis. If an 
        atom has an atomic charge less than :attr:`max_charge`, and it is 
        within :attr:`d_h_cutoff` of a hydrogen atom, then it is considered 
        capable of participating in hydrogen bonds.

        If :attr:`donors_sel` is `None`, and the universe topology does not 
        have bonding information, this function is called to guess the 
        selection.

        Alternatively, this function may be used to quickly generate a 
        :class:`str` of potential donor atoms involved in hydrogen bonding. 
        This :class:`str` may then be modified before being used to set the 
        attribute :attr:`donors_sel`.


        .. versionchanged:: 2.4.0
            Added ability to use atom types

        """

        # We need to know `hydrogens_sel` before we can find donors
        # Use a new variable `hydrogens_sel` so that we do not set 
        # `self.hydrogens_sel` if it is currently `None`
        if self.hydrogens_sel is None:
            hydrogens_sel = self.guess_hydrogens()
        else:
            hydrogens_sel = self.hydrogens_sel
        hydrogens_ag = self.u.select_atoms(hydrogens_sel)

        # We're using u._topology.bonds rather than u.bonds as it is a million
        # times faster to access. This is because u.bonds also calculates
        # properties of each bond (e.g bond length). See:
        # https://github.com/MDAnalysis/mdanalysis/issues/2396#issuecomment-596251787
        if (hasattr(self.u._topology, 'bonds') 
           and len(self.u._topology.bonds.values) != 0):
            donors_ag = find_hydrogen_donors(hydrogens_ag)
            donors_ag = donors_ag.intersection(self.u.select_atoms(select))
        else:
            donors_ag = hydrogens_ag.residues.atoms.select_atoms(
                "({donors_sel}) and around {d_h_cutoff} {hydrogens_sel}".format(
                    donors_sel=select,
                    d_h_cutoff=self.d_h_cutoff,
                    hydrogens_sel=hydrogens_sel
                )
            )

        donors_ag = donors_ag[donors_ag.charges < max_charge]

        return self._group_categories(donors_ag)

    def guess_acceptors(self, select='all', max_charge=-0.5):
        """Guesses which atoms could be considered acceptors in the analysis.

        Acceptor selections may be achieved with either a resname, atom 
        name combination, or when those are absent, atom types.

        Parameters
        ----------
        select: str (optional)
            :ref:`Selection string <selection-commands-label>` for atom group
            from which acceptors will be identified. (e.g., ``(resname X and 
            name O1)`` or ``type 2``)
        max_charge: float (optional)
            The charge of an acceptor atom must be less than this value.

        Returns
        -------
        potential_acceptors: str
            String containing the :attr:`resname` and :attr:`name` of all atoms 
            that potentially capable of forming hydrogen bonds.

        Notes
        -----
        Acceptor selections may be achieved with either a resname, atom 
        name combination, or when those are absent, atom types.

        This function makes use of and atomic charges to identify which atoms 
        could be considered acceptor atoms in the hydrogen bond analysis. If 
        an atom has an atomic charge less than :attr:`max_charge` then it is 
        considered capable of participating in hydrogen bonds.

        If :attr:`acceptors_sel` is `None`, this function is called to guess 
        the selection.

        Alternatively, this function may be used to quickly generate a 
        :class:`str` of potential acceptor atoms involved in hydrogen bonding. 
        This :class:`str` may then be modified before being used to set the 
        attribute :attr:`acceptors_sel`.


        .. versionchanged:: 2.4.0
            Added ability to use atom types

        """

        ag = self.u.select_atoms(select)
        acceptors_ag = ag[ag.charges < max_charge]

        return self._group_categories(acceptors_ag)

    @staticmethod
    def _group_categories(group):
        """ Find categories according to universe constraints
        
        Parameters
        ----------
        group : AtomGroup
            AtomGroups corresponding to either hydrogen bond acceptors, 
            donors, or hydrogen atoms that meet their respective charge
            and mass constraints. 

        Returns
        -------
        select : str
            String for each hydrogen bond acceptor/donor/hydrogen atom category.


        .. versionadded:: 2.4.0

        """

        if hasattr(group, "resnames") and hasattr(group, "names"):
            group_list = np.unique([
                '(resname {} and name {})'.format(r, 
                    p) for r, p in zip(group.resnames, group.names)
            ])
        else:
            group_list = np.unique(
                [
                    'type {}'.format(tp) for tp in group.types
                ]
            )

        return " or ".join(group_list)

    def _get_dh_pairs(self):
        """Finds donor-hydrogen pairs.

        Returns
        -------
        donors, hydrogens: AtomGroup, AtomGroup
            AtomGroups corresponding to all donors and all hydrogens. 
            AtomGroups are ordered such that, if zipped, will
            produce a list of donor-hydrogen pairs.
        """

        # If donors_sel is not provided, use topology to find d-h pairs
        if self.donors_sel is None:

            # We're using u._topology.bonds rather than u.bonds as it is a million times faster to access.
            # This is because u.bonds also calculates properties of each bond (e.g bond length).
            # See https://github.com/MDAnalysis/mdanalysis/issues/2396#issuecomment-596251787
            if not (hasattr(self.u._topology, 'bonds') and len(self.u._topology.bonds.values) != 0):
                raise NoDataError('Cannot assign donor-hydrogen pairs via topology as no bond information is present. '
                                  'Please either: load a topology file with bond information; use the guess_bonds() '
                                  'topology guesser; or set HydrogenBondAnalysis.donors_sel so that a distance cutoff '
                                  'can be used.')

            hydrogens = self.u.select_atoms(self.hydrogens_sel)
            donors = sum(h.bonded_atoms[0] for h in hydrogens) if hydrogens \
                else AtomGroup([], self.u)

        # Otherwise, use d_h_cutoff as a cutoff distance
        else:

            hydrogens = self.u.select_atoms(self.hydrogens_sel)
            donors = self.u.select_atoms(self.donors_sel)
            donors_indices, hydrogen_indices = capped_distance(
                donors.positions,
                hydrogens.positions,
                max_cutoff=self.d_h_cutoff,
                box=self.u.dimensions,
                return_distances=False
            ).T

            donors = donors[donors_indices]
            hydrogens = hydrogens[hydrogen_indices]

        return donors, hydrogens

    def _filter_atoms(self, donors, acceptors):
        """Create a mask to filter donor, hydrogen and acceptor atoms.

        This can be used to consider only hydrogen bonds between two or more
        specified groups.

        Groups are specified with the `between` keyword when creating the
        HydrogenBondAnalysis object.

           Returns
           -------
           mask: np.ndarray


        .. versionchanged:: 2.5.0
           Change return value to a mask instead of separate AtomGroups.
``
        """

        mask = np.full(donors.n_atoms, fill_value=False)
        for group1, group2 in self.between_ags:

            # Find donors in G1 and acceptors in G2
            mask[
                    np.logical_and(
                        np.isin(donors.indices, group1.indices),
                        np.isin(acceptors.indices, group2.indices)
                    )
            ] = True

            # Find acceptors in G1 and donors in G2
            mask[
                np.logical_and(
                    np.isin(acceptors.indices, group1.indices),
                    np.isin(donors.indices, group2.indices)
                )
            ] = True

        return mask


    def _prepare(self):
        self.results.hbonds = [[], [], [], [], [], []]


    def _single_frame(self):

        box = self._ts.dimensions

        # Update donor-hydrogen pairs if necessary
        if self.update_selections:
            self._donors, self._hydrogens = self._get_dh_pairs()

        # find D and A within cutoff distance of one another
        # min_cutoff = 1.0 as an atom cannot form a hydrogen bond with itself
        d_a_indices, d_a_distances = capped_distance(
            self._donors.positions,
            self._acceptors.positions,
            max_cutoff=self.d_a_cutoff,
            min_cutoff=1.0,
            box=box,
            return_distances=True,
        )

        if np.size(d_a_indices) == 0:
            warnings.warn(
                "No hydrogen bonds were found given d-a cutoff of "
                f"{self.d_a_cutoff} between Donor, {self.donors_sel}, and "
                f"Acceptor, {self.acceptors_sel}."
            )

        # Remove D-A pairs more than d_a_cutoff away from one another
        tmp_donors = self._donors[d_a_indices.T[0]]
        tmp_hydrogens = self._hydrogens[d_a_indices.T[0]]
        tmp_acceptors = self._acceptors[d_a_indices.T[1]]

        # Remove donor-acceptor pairs between pairs of AtomGroups we are not
        # interested in
        if self.between_ags is not None:
            between_mask = self._filter_atoms(tmp_donors, tmp_acceptors)
            tmp_donors = tmp_donors[between_mask]
            tmp_hydrogens = tmp_hydrogens[between_mask]
            tmp_acceptors = tmp_acceptors[between_mask]
            d_a_distances = d_a_distances[between_mask]

        # Find D-H-A angles greater than d_h_a_angle_cutoff
        d_h_a_angles = np.rad2deg(
            calc_angles(
                tmp_donors.positions,
                tmp_hydrogens.positions,
                tmp_acceptors.positions,
                box=box
            )
        )
        hbond_indices = np.where(d_h_a_angles > self.d_h_a_angle)[0]

        if np.size(hbond_indices) == 0:
            warnings.warn(
                "No hydrogen bonds were found given angle of "
                f"{self.d_h_a_angle} between Donor, {self.donors_sel}, and "
                f"Acceptor, {self.acceptors_sel}."
            )

        # Retrieve atoms, distances and angles of hydrogen bonds
        hbond_donors = tmp_donors[hbond_indices]
        hbond_hydrogens = tmp_hydrogens[hbond_indices]
        hbond_acceptors = tmp_acceptors[hbond_indices]
        hbond_distances = d_a_distances[hbond_indices]
        hbond_angles = d_h_a_angles[hbond_indices]

        # Store data on hydrogen bonds found at this frame
        self.results.hbonds[0].extend(np.full_like(hbond_donors,
                                      self._ts.frame))
        self.results.hbonds[1].extend(hbond_donors.indices)
        self.results.hbonds[2].extend(hbond_hydrogens.indices)
        self.results.hbonds[3].extend(hbond_acceptors.indices)
        self.results.hbonds[4].extend(hbond_distances)
        self.results.hbonds[5].extend(hbond_angles)

    def _conclude(self):

        self.results.hbonds = np.asarray(self.results.hbonds).T

    def _get_aggregator(self):
        return ResultsGroup(lookup={'hbonds': ResultsGroup.ndarray_hstack})

    @property
    def hbonds(self):
        wmsg = ("The `hbonds` attribute was deprecated in MDAnalysis 2.0.0 "
                "and will be removed in MDAnalysis 3.0.0. Please use "
                "`results.hbonds` instead.")
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.hbonds

    def lifetime(self, tau_max=20, window_step=1, intermittency=0):
        """Computes and returns the time-autocorrelation
        (HydrogenBondLifetimes) of hydrogen bonds.

        Before calling this method, the hydrogen bonds must first be computed
        with the `run()` function. The same `start`, `stop` and `step`
        parameters used in finding hydrogen bonds will be used here for
        calculating hydrogen bond lifetimes. That is, the same frames will be
        used in the analysis.

        Unique hydrogen bonds are identified using hydrogen-acceptor pairs.
        This means an acceptor switching to a different hydrogen atom - with
        the same donor - from one frame to the next is considered a different
        hydrogen bond.

        Please see :func:`MDAnalysis.lib.correlations.autocorrelation` and
        :func:`MDAnalysis.lib.correlations.intermittency` functions for more
        details.


        Parameters
        ----------
        window_step : int, optional
            The number of frames between each t(0).
        tau_max : int, optional
            Hydrogen bond lifetime is calculated for frames in the range
            1 <= `tau` <= `tau_max`
        intermittency : int, optional
            The maximum number of consecutive frames for which a bond can
            disappear but be counted as present if it returns at the next
            frame. An intermittency of `0` is equivalent to a continuous
            autocorrelation, which does not allow for hydrogen bond
            disappearance. For example, for `intermittency=2`, any given
            hydrogen bond may disappear for up to two consecutive frames yet
            be treated as being present at all frames. The default is
            continuous (intermittency=0).

        Returns
        -------
        tau_timeseries : np.array
            tau from 1 to `tau_max`
        timeseries : np.array
            autcorrelation value for each value of `tau`
        """

        if self.results.hbonds is None:
            logger.error(
                "Autocorrelation analysis of hydrogen bonds cannot be done"
                "before the hydrogen bonds are found"
            )
            logger.error(
                "Autocorrelation: Please use the .run() before calling this"
                "function"
            )
            raise NoDataError(".hbonds attribute is None: use .run() first")

        if self.step != 1:
            logger.warning(
                "Autocorrelation: Hydrogen bonds were computed with step > 1."
            )
            logger.warning(
                "Autocorrelation: We recommend recomputing hydrogen bonds with"
                " step = 1."
            )
            logger.warning(
                "Autocorrelation: if you would like to allow bonds to break"
                " and reform, please use 'intermittency'"
            )

        # Extract the hydrogen bonds IDs only in the format
        # [set(superset(x1,x2), superset(x3,x4)), ..]
        found_hydrogen_bonds = [set() for _ in self.frames]
        for frame_index, frame in enumerate(self.frames):
            for hbond in self.results.hbonds[self.results.hbonds[:, 0] == frame]:
                found_hydrogen_bonds[frame_index].add(frozenset(hbond[2:4]))

        intermittent_hbonds = correct_intermittency(
            found_hydrogen_bonds,
            intermittency=intermittency
        )
        tau_timeseries, timeseries, timeseries_data = autocorrelation(
            intermittent_hbonds,
            tau_max,
            window_step=window_step
        )

        return np.vstack([tau_timeseries, timeseries])

    def count_by_time(self):
        """Counts the number of hydrogen bonds per timestep.

        Returns
        -------
        counts : numpy.ndarray
             Contains the total number of hydrogen bonds found at each timestep.
             Can be used along with :attr:`HydrogenBondAnalysis.times` to plot
             the number of hydrogen bonds over time.
        """

        indices, tmp_counts = np.unique(self.results.hbonds[:, 0], axis=0,
                                        return_counts=True)

        indices -= self.start
        indices /= self.step

        counts = np.zeros_like(self.frames)
        counts[indices.astype(np.intp)] = tmp_counts

        return counts

    def count_by_type(self):
        """Counts the total number of each unique type of hydrogen bond.

        Returns
        -------
        counts : numpy.ndarray
             Each row of the array contains the donor resname, donor atom type, 
             acceptor resname, acceptor atom type and the total number of times 
             the hydrogen bond was found.

        Note
        ----
        Unique hydrogen bonds are determined through a consideration of the 
        resname and atom type of the donor and acceptor atoms in a hydrogen bond.
        """

        d = self.u.atoms[self.results.hbonds[:, 1].astype(np.intp)]
        a = self.u.atoms[self.results.hbonds[:, 3].astype(np.intp)]

        if hasattr(d, "resnames"):
            d_res = d.resnames
            a_res = a.resnames
        else:
            d_res = len(d.types) * ["None"]
            a_res = len(a.types) * ["None"]

        tmp_hbonds = np.array([d_res, d.types, a_res, a.types], dtype=str).T
        hbond_type, type_counts = np.unique(
            tmp_hbonds, axis=0, return_counts=True)
        hbond_type_list = []
        for hb_type, hb_count in zip(hbond_type, type_counts):
            hbond_type_list.append([":".join(hb_type[:2]),
                                    ":".join(hb_type[2:4]), hb_count])

        return np.array(hbond_type_list)

    def count_by_ids(self):
        """Counts the total number hydrogen bonds formed by unique combinations of donor, hydrogen and acceptor atoms.

        Returns
        -------
        counts : numpy.ndarray
             Each row of the array contains the donor atom id, hydrogen atom id, acceptor atom id and the total number
             of times the hydrogen bond was observed. The array is sorted by frequency of occurrence.

        Note
        ----
        Unique hydrogen bonds are determined through a consideration of the hydrogen atom id and acceptor atom id
        in a hydrogen bond.
        """

        d = self.u.atoms[self.results.hbonds[:, 1].astype(np.intp)]
        h = self.u.atoms[self.results.hbonds[:, 2].astype(np.intp)]
        a = self.u.atoms[self.results.hbonds[:, 3].astype(np.intp)]

        tmp_hbonds = np.array([d.ids, h.ids, a.ids]).T
        hbond_ids, ids_counts = np.unique(tmp_hbonds, axis=0,
                                          return_counts=True)

        # Find unique hbonds and sort rows so that most frequent observed bonds are at the top of the array
        unique_hbonds = np.concatenate((hbond_ids, ids_counts[:, None]),
                                       axis=1)
        unique_hbonds = unique_hbonds[unique_hbonds[:, 3].argsort()[::-1]]

        return unique_hbonds

    @property
    def donors_sel(self):
        """Selection string for the hydrogen bond donor atoms.
        
        .. versionadded:: 2.10.0
        """
        return self._donors_sel

    @donors_sel.setter
    def donors_sel(self, value):
        self._donors_sel = value
        self._donors, self._hydrogens = self._get_dh_pairs()

    @property
    def hydrogens_sel(self):
        """Selection string for the hydrogen bond hydrogen atoms.
        
        .. versionadded:: 2.10.0
        """
        return self._hydrogens_sel

    @hydrogens_sel.setter
    def hydrogens_sel(self, value):
        self._hydrogens_sel = value
        if self._hydrogens_sel is None:
            self._hydrogens_sel = self.guess_hydrogens()
        self._donors, self._hydrogens = self._get_dh_pairs()

    @property
    def acceptors_sel(self):
        """Selection string for the hydrogen bond acceptor atoms.
        
        .. versionadded:: 2.10.0
        """
        return self._acceptors_sel

    @acceptors_sel.setter
    def acceptors_sel(self, value):
        self._acceptors_sel = value
        if self._acceptors_sel is None:
            self._acceptors_sel = self.guess_acceptors()
        self._acceptors = self.u.select_atoms(self._acceptors_sel,
                                              updating=self.update_selections)


# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Version information for MDAnalysis - :mod:`MDAnalysis.version`
==============================================================

The version information in :mod:`MDAnalysis.version` indicates the
release of MDAnalysis. MDAnalysis uses `semantic versioning`_ as
described in the wiki page on `versioning of MDAnalysis`_.

In brief:

Given a version number MAJOR.MINOR.PATCH, we increment the

1. **MAJOR** version when we make **incompatible API changes**,
2. **MINOR** version when we **add functionality** in a
   **backwards-compatible** manner, and
3. **PATCH** version when we make backwards-compatible **bug fixes**.

However, as long as the **MAJOR** number is **0** (i.e. the API has
not stabilized), even **MINOR** increases *may* introduce incompatible
API changes. As soon as we have a 1.0.0 release, the public API can
only be changed in a backward-incompatible manner with an increase in
MAJOR version.

Additional labels for pre-release and build metadata are available as
extensions to the MAJOR.MINOR.PATCH format.


.. Note:: Development versions and pre-releases have a suffix after
          the release number, such as ``0.11.0-dev``. If you have
          problems, try out a full release (e.g. ``0.11.0``) first.

.. _`semantic versioning`: http://semver.org/
.. _`versioning of MDAnalysis`:
   https://github.com/MDAnalysis/mdanalysis/wiki/SemanticVersioning

Data
----

.. autodata:: __version__

"""

# keep __version__ in separate file to avoid circular imports
# e.g. with lib.log

#: Release of MDAnalysis as a string, using `semantic versioning`_.
__version__ = "2.10.0"  # NOTE: keep in sync with RELEASE in setup.py



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
import pytest
import copy
import sys
import warnings
from unittest.mock import Mock, patch

from numpy.testing import assert_equal, assert_almost_equal

import gridData.OpenDX

import MDAnalysis as mda
from MDAnalysis.analysis import density

from MDAnalysisTests.datafiles import TPR, XTC
from MDAnalysisTests.util import block_import


class TestDensity(object):
    nbins = 3, 4, 5
    counts = 100
    Lmax = 10.0

    @pytest.fixture(scope="class")
    def bins(self):
        return [np.linspace(0, self.Lmax, n + 1) for n in self.nbins]

    @pytest.fixture()
    def h_and_edges(self, bins):
        return np.histogramdd(
            self.Lmax
            * np.sin(np.linspace(0, 1, self.counts * 3)).reshape(
                self.counts, 3
            ),
            bins=bins,
        )

    @pytest.fixture()
    def D(self, h_and_edges):
        h, edges = h_and_edges
        d = density.Density(
            h, edges, parameters={"isDensity": False}, units={"length": "A"}
        )
        d.make_density()
        return d

    @pytest.fixture()
    def D1(self, h_and_edges):
        h, edges = h_and_edges
        d = density.Density(h, edges, parameters={}, units={})
        return d

    def test_shape(self, D):
        assert D.grid.shape == self.nbins

    def test_edges(self, bins, D):
        for dim, (edges, fixture) in enumerate(zip(D.edges, bins)):
            assert_almost_equal(
                edges, fixture, err_msg="edges[{0}] mismatch".format(dim)
            )

    def test_midpoints(self, bins, D):
        midpoints = [0.5 * (b[:-1] + b[1:]) for b in bins]
        for dim, (mp, fixture) in enumerate(zip(D.midpoints, midpoints)):
            assert_almost_equal(
                mp, fixture, err_msg="midpoints[{0}] mismatch".format(dim)
            )

    def test_delta(self, D):
        deltas = np.array([self.Lmax]) / np.array(self.nbins)
        assert_almost_equal(D.delta, deltas)

    def test_grid(self, D):
        dV = D.delta.prod()  # orthorhombic grids only!
        # counts = (rho[0] * dV[0] + rho[1] * dV[1] ...) = sum_i rho[i] * dV
        assert_almost_equal(D.grid.sum() * dV, self.counts)

    def test_origin(self, bins, D):
        midpoints = [0.5 * (b[:-1] + b[1:]) for b in bins]
        origin = [m[0] for m in midpoints]
        assert_almost_equal(D.origin, origin)

    def test_check_set_unit_keyerror(self, D):
        units = {"weight": "A"}
        with pytest.raises(ValueError):
            D._check_set_unit(units)

    def test_check_set_unit_attributeError(self, D):
        units = []
        with pytest.raises(ValueError):
            D._check_set_unit(units)

    def test_check_set_unit_nolength(self, D):
        del D.units["length"]
        units = {"density": "A^{-3}"}
        with pytest.raises(ValueError):
            D._check_set_unit(units)

    def test_check_set_density_none(self, D1):
        units = {"density": None}
        D1._check_set_unit(units)
        assert D1.units["density"] is None

    def test_check_set_density_not_in_units(self, D1):
        del D1.units["density"]
        D1._check_set_unit({})
        assert D1.units["density"] is None

    def test_parameters_isdensity(self, D):
        with pytest.warns(UserWarning, match="Running make_density()"):
            D.make_density()

    def test_check_convert_density_grid_not_density(self, D1):
        with pytest.raises(RuntimeError, match="The grid is not a density"):
            D1.convert_density()

    def test_check_convert_density_value_error(self, D):
        unit = "A^{-2}"
        with pytest.raises(ValueError, match="The name of the unit"):
            D.convert_density(unit)

    def test_check_convert_density_units_same_density_units(self, D):
        unit = "A^{-3}"
        D_orig = copy.deepcopy(D)
        D.convert_density(unit)
        assert D.units["density"] == D_orig.units["density"] == unit
        assert_almost_equal(D.grid, D_orig.grid)

    def test_check_convert_density_units_density(self, D):
        unit = "nm^{-3}"
        D_orig = copy.deepcopy(D)
        D.convert_density(unit)
        assert D.units["density"] == "nm^{-3}"
        assert_almost_equal(D.grid, 10**3 * D_orig.grid)

    def test_convert_length_same_length_units(self, D):
        unit = "A"
        D_orig = copy.deepcopy(D)
        D.convert_length(unit)
        assert D.units["length"] == D_orig.units["length"] == unit
        assert_almost_equal(D.grid, D_orig.grid)

    def test_convert_length_other_length_units(self, D):
        unit = "nm"
        D_orig = copy.deepcopy(D)
        D.convert_length(unit)
        assert D.units["length"] == unit
        assert_almost_equal(D.grid, D_orig.grid)

    def test_repr(self, D, D1):
        assert str(D) == "<Density density with (3, 4, 5) bins>"
        assert str(D1) == "<Density histogram with (3, 4, 5) bins>"

    def test_check_convert_length_edges(self, D):
        D1 = copy.deepcopy(D)
        unit = "nm"
        D.convert_length(unit)
        for prev_edge, conv_edge in zip(D1.edges, D.edges):
            assert_almost_equal(prev_edge, 10 * conv_edge)

    def test_check_convert_density_edges(self, D):
        unit = "nm^{-3}"
        D_orig = copy.deepcopy(D)
        D.convert_density(unit)
        for new_den, orig_den in zip(D.edges, D_orig.edges):
            assert_almost_equal(new_den, orig_den)

    @pytest.mark.parametrize("dxtype", ("float", "double", "int", "byte"))
    def test_export_types(self, D, dxtype, tmpdir, outfile="density.dx"):
        with tmpdir.as_cwd():
            D.export(outfile, type=dxtype)

            dx = gridData.OpenDX.field(0)
            dx.read(outfile)
            data = dx.components["data"]
        assert data.type == dxtype


class DensityParameters(object):
    topology = TPR
    trajectory = XTC
    delta = 2.0
    selections = {
        "none": "resname None",
        "static": "name OW",
        "dynamic": "name OW and around 4 (protein and resid 1-10)",
        "solute": "protein and not name H*",
    }
    references = {
        "static": {
            "meandensity": 0.016764271713091212,
        },
        "static_sliced": {
            "meandensity": 0.016764270747693617,
        },
        "static_defined": {
            "meandensity": 0.0025000000000000005,
        },
        "static_defined_unequal": {
            "meandensity": 0.006125,
        },
        "dynamic": {
            "meandensity": 0.0012063418843728784,
        },
        "notwithin": {
            "meandensity": 0.015535385132107926,
        },
    }
    cutoffs = {
        "notwithin": 4.0,
    }
    gridcenters = {
        "static_defined": np.array([56.0, 45.0, 35.0]),
        "error1": np.array([56.0, 45.0]),
        "error2": [56.0, 45.0, "MDAnalysis"],
    }
    precision = 5
    outfile = "density.dx"

    @pytest.fixture()
    def universe(self):
        return mda.Universe(
            self.topology, self.trajectory, tpr_resid_from_one=False
        )


class TestDensityAnalysis(DensityParameters):
    def check_DensityAnalysis(
        self,
        ag,
        ref_meandensity,
        tmpdir,
        client_DensityAnalysis,
        runargs=None,
        **kwargs,
    ):
        runargs = runargs if runargs else {}
        with tmpdir.as_cwd():
            D = density.DensityAnalysis(ag, delta=self.delta, **kwargs).run(
                **runargs, **client_DensityAnalysis
            )
            assert_almost_equal(
                D.results.density.grid.mean(),
                ref_meandensity,
                err_msg="mean density does not match",
            )
            D.results.density.export(self.outfile)

            D2 = density.Density(self.outfile)
            assert_almost_equal(
                D.results.density.grid,
                D2.grid,
                decimal=self.precision,
                err_msg="DX export failed: different grid sizes",
            )

    @pytest.mark.parametrize("mode", ("static", "dynamic"))
    def test_run(self, mode, universe, tmpdir, client_DensityAnalysis):
        updating = mode == "dynamic"
        self.check_DensityAnalysis(
            universe.select_atoms(self.selections[mode], updating=updating),
            self.references[mode]["meandensity"],
            tmpdir=tmpdir,
            client_DensityAnalysis=client_DensityAnalysis,
        )

    def test_sliced(self, universe, tmpdir, client_DensityAnalysis):
        self.check_DensityAnalysis(
            universe.select_atoms(self.selections["static"]),
            self.references["static_sliced"]["meandensity"],
            tmpdir=tmpdir,
            client_DensityAnalysis=client_DensityAnalysis,
            runargs=dict(start=1, stop=-1, step=2),
        )

    def test_userdefn_eqbox(self, universe, tmpdir, client_DensityAnalysis):
        with warnings.catch_warnings():
            # Do not need to see UserWarning that box is too small
            warnings.simplefilter("ignore")
            self.check_DensityAnalysis(
                universe.select_atoms(self.selections["static"]),
                self.references["static_defined"]["meandensity"],
                tmpdir=tmpdir,
                client_DensityAnalysis=client_DensityAnalysis,
                gridcenter=self.gridcenters["static_defined"],
                xdim=10.0,
                ydim=10.0,
                zdim=10.0,
            )

    def test_userdefn_neqbox(self, universe, tmpdir, client_DensityAnalysis):
        self.check_DensityAnalysis(
            universe.select_atoms(self.selections["static"]),
            self.references["static_defined_unequal"]["meandensity"],
            tmpdir=tmpdir,
            client_DensityAnalysis=client_DensityAnalysis,
            gridcenter=self.gridcenters["static_defined"],
            xdim=10.0,
            ydim=15.0,
            zdim=20.0,
        )

    def test_userdefn_boxshape(self, universe, client_DensityAnalysis):
        D = density.DensityAnalysis(
            universe.select_atoms(self.selections["static"]),
            delta=1.0,
            xdim=8.0,
            ydim=12.0,
            zdim=17.0,
            gridcenter=self.gridcenters["static_defined"],
        ).run(**client_DensityAnalysis)
        assert D.results.density.grid.shape == (8, 12, 17)

    def test_warn_userdefn_padding(self, universe, client_DensityAnalysis):
        regex = (
            r"Box padding \(currently set at 1\.0\) is not used "
            r"in user defined grids\."
        )
        with pytest.warns(UserWarning, match=regex):
            D = density.DensityAnalysis(
                universe.select_atoms(self.selections["static"]),
                delta=self.delta,
                xdim=100.0,
                ydim=100.0,
                zdim=100.0,
                padding=1.0,
                gridcenter=self.gridcenters["static_defined"],
            ).run(step=5, **client_DensityAnalysis)

    def test_warn_userdefn_smallgrid(self, universe, client_DensityAnalysis):
        regex = (
            "Atom selection does not fit grid --- "
            "you may want to define a larger box"
        )
        with pytest.warns(UserWarning, match=regex):
            D = density.DensityAnalysis(
                universe.select_atoms(self.selections["static"]),
                delta=self.delta,
                xdim=1.0,
                ydim=2.0,
                zdim=2.0,
                padding=0.0,
                gridcenter=self.gridcenters["static_defined"],
            ).run(step=5, **client_DensityAnalysis)

    def test_ValueError_userdefn_gridcenter_shape(
        self, universe, client_DensityAnalysis
    ):
        # Test len(gridcenter) != 3
        with pytest.raises(
            ValueError, match="Gridcenter must be a 3D coordinate"
        ):
            D = density.DensityAnalysis(
                universe.select_atoms(self.selections["static"]),
                delta=self.delta,
                xdim=10.0,
                ydim=10.0,
                zdim=10.0,
                gridcenter=self.gridcenters["error1"],
            ).run(step=5, **client_DensityAnalysis)

    def test_ValueError_userdefn_gridcenter_type(
        self, universe, client_DensityAnalysis
    ):
        # Test gridcenter includes non-numeric strings
        with pytest.raises(
            ValueError, match="Gridcenter must be a 3D coordinate"
        ):
            D = density.DensityAnalysis(
                universe.select_atoms(self.selections["static"]),
                delta=self.delta,
                xdim=10.0,
                ydim=10.0,
                zdim=10.0,
                gridcenter=self.gridcenters["error2"],
            ).run(step=5, **client_DensityAnalysis)

    def test_ValueError_userdefn_gridcenter_missing(
        self, universe, client_DensityAnalysis
    ):
        # Test no gridcenter provided when grid dimensions are given
        regex = "Gridcenter or grid dimensions are not provided"
        with pytest.raises(ValueError, match=regex):
            D = density.DensityAnalysis(
                universe.select_atoms(self.selections["static"]),
                delta=self.delta,
                xdim=10.0,
                ydim=10.0,
                zdim=10.0,
            ).run(step=5, **client_DensityAnalysis)

    def test_ValueError_userdefn_xdim_type(
        self, universe, client_DensityAnalysis
    ):
        # Test xdim != int or float
        with pytest.raises(
            ValueError, match="xdim, ydim, and zdim must be numbers"
        ):
            D = density.DensityAnalysis(
                universe.select_atoms(self.selections["static"]),
                delta=self.delta,
                xdim="MDAnalysis",
                ydim=10.0,
                zdim=10.0,
                gridcenter=self.gridcenters["static_defined"],
            ).run(step=5, **client_DensityAnalysis)

    def test_ValueError_userdefn_xdim_nanvalue(
        self, universe, client_DensityAnalysis
    ):
        # Test  xdim set to NaN value
        regex = "Gridcenter or grid dimensions have NaN element"
        with pytest.raises(ValueError, match=regex):
            D = density.DensityAnalysis(
                universe.select_atoms(self.selections["static"]),
                delta=self.delta,
                xdim=np.nan,
                ydim=10.0,
                zdim=10.0,
                gridcenter=self.gridcenters["static_defined"],
            ).run(step=5, **client_DensityAnalysis)

    def test_warn_noatomgroup(self, universe, client_DensityAnalysis):
        regex = (
            "No atoms in AtomGroup at input time frame. "
            "This may be intended; please ensure that "
            "your grid selection covers the atomic "
            "positions you wish to capture."
        )
        with pytest.warns(UserWarning, match=regex):
            D = density.DensityAnalysis(
                universe.select_atoms(self.selections["none"]),
                delta=self.delta,
                xdim=1.0,
                ydim=2.0,
                zdim=2.0,
                padding=0.0,
                gridcenter=self.gridcenters["static_defined"],
            ).run(step=5, **client_DensityAnalysis)

    def test_ValueError_noatomgroup(self, universe, client_DensityAnalysis):
        with pytest.raises(
            ValueError,
            match="No atoms in AtomGroup at input"
            " time frame. Grid for density"
            " could not be automatically"
            " generated. If this is"
            " expected, a user"
            " defined grid will "
            "need to be provided instead.",
        ):
            D = density.DensityAnalysis(
                universe.select_atoms(self.selections["none"])
            ).run(step=5, **client_DensityAnalysis)

    def test_warn_results_deprecated(self, universe, client_DensityAnalysis):
        D = density.DensityAnalysis(
            universe.select_atoms(self.selections["static"])
        )
        D.run(stop=1, **client_DensityAnalysis)
        wmsg = "The `density` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(D.density.grid, D.results.density.grid)

    def test_density_analysis_conversion_default_unit(self):
        u = mda.Universe(TPR, XTC)
        ow = u.select_atoms("name OW")
        D = mda.analysis.density.DensityAnalysis(ow, delta=1.0)
        D.run()
        D.results.density.convert_density()


class TestGridImport(object):

    @block_import("gridData")
    def test_absence_griddata(self):
        sys.modules.pop("MDAnalysis.analysis.density", None)
        # if gridData package is missing an ImportError should be raised
        # at the module level of MDAnalysis.analysis.density
        with pytest.raises(ImportError):
            import MDAnalysis.analysis.density

    def test_presence_griddata(self):
        sys.modules.pop("MDAnalysis.analysis.density", None)
        # no ImportError exception is raised when gridData is properly
        # imported by MDAnalysis.analysis.density

        # mock gridData in case there are testing scenarios where
        # it is not available
        mock = Mock()
        with patch.dict("sys.modules", {"gridData": mock}):
            try:
                import MDAnalysis.analysis.density
            except ImportError:
                pytest.fail(
                    msg="""MDAnalysis.analysis.density should not raise
                             an ImportError if gridData is available."""
                )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""
Common functions for coordinate reading --- :mod:`MDAnalysis.coordinates.core`
==============================================================================

Important base classes are collected in :mod:`MDAnalysis.coordinates.base`.

.. autofunction:: reader
.. autofunction:: writer

Helper functions:

.. autofunction:: get_reader_for
.. autofunction:: get_writer_for

"""

from ..lib import util
from ..lib.mdamath import triclinic_box, triclinic_vectors, box_volume
from ..core._get_readers import get_reader_for, get_writer_for


def reader(filename, format=None, **kwargs):
    """Provide a trajectory reader instance for *filename*.

    This function guesses the file format from the extension of *filename* and
    it will throw a :exc:`TypeError` if the extension is not recognized.

    In most cases, no special keyword arguments are necessary.

    All other keywords are passed on to the underlying Reader classes; see
    their documentation for details.

    Parameters
    ----------
    filename : str or tuple
        filename (or tuple of filenames) of the input coordinate file
    kwargs
        Keyword arguments for the selected Reader class.

    Returns
    -------
    :class:`~base.Reader`
        A trajectory Reader instance

    See Also
    --------
    :ref:`Supported coordinate formats`


    """
    if isinstance(filename, tuple):
        Reader = get_reader_for(filename[0], format=filename[1])
        filename = filename[0]
    else:
        Reader = get_reader_for(filename, format=format)
    try:
        return Reader(filename, **kwargs)
    except ValueError:
        errmsg = f"Unable to read {filename} with {Reader}."
        raise TypeError(errmsg) from None


def writer(filename, n_atoms=None, **kwargs):
    """Initialize a trajectory writer instance for *filename*.

    Parameters
    ----------
    filename : str
        Output filename of the trajectory; the extension determines the
        format.
    n_atoms : int (optional)
        The number of atoms in the output trajectory; can be ommitted
        for single-frame writers.
    multiframe : bool (optional)
        ``True``: write a trajectory with multiple frames; ``False``
        only write a single frame snapshot; ``None`` first try to get
        a multiframe writer and then fall back to single frame [``None``]
    kwargs : optional
        Keyword arguments for the writer; all trajectory Writers accept
        ``start``: starting time [0], ``step``: step size in frames [1],
        ``dt``: length of time between two frames, in ps [1.0] Some readers
        accept additional arguments, which need to be looked up in the
        documentation of the reader.

    Returns
    -------
    :class:`~base.Writer`
        A trajectory Writer instance

    See Also
    --------
    :ref:`Supported coordinate formats`


    .. versionchanged:: 0.7.6
       Added `multiframe` keyword. See also :func:`get_writer_for`.

    """
    Writer = get_writer_for(
        filename,
        format=kwargs.pop("format", None),
        multiframe=kwargs.pop("multiframe", None),
    )
    return Writer(filename, n_atoms=n_atoms, **kwargs)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
#

import cython
import numpy as np
from .mdamath import triclinic_vectors
cimport numpy as cnp
cimport MDAnalysis.lib._cutil
from MDAnalysis.lib._cutil cimport _dot, _norm, _cross

from libcpp.vector cimport vector

cnp.import_array()


__all__ = ['augment_coordinates', 'undo_augment']


@cython.boundscheck(False)
@cython.wraparound(False)
def augment_coordinates(float[:, ::1] coordinates, float[:] box, float r):
    r"""Calculates the periodic images of particles which are within a distance
    `r` from the box walls.

    The algorithm works by generating explicit periodic images of atoms residing
    close to any of the six box walls. The steps involved in generating images
    involves the evaluation of reciprocal box vectors followed by the
    calculation of distances of atoms from the walls by means of projection onto
    the reciprocal vectors. If the distance is less than a specified cutoff
    distance, relevant periodic images are generated using box translation
    vectors :math:`\vec{t}` with

    .. math:: \vec{t}=l\cdot\vec{a}+m\cdot\vec{b}+n\cdot \vec{c}\,,

    where :math:`l,\,m,\,n \in \{-1,\,0,\,1\}` are the neighboring cell indices
    in :math:`x`-, :math:`y`-, and :math:`z`-direction relative to the central
    cell with box vectors :math:`\vec{a},\,\vec{b},\,\vec{c}`.

    For instance, an atom close to the :math:`xy`-plane containing the origin
    will generate a periodic image outside the central cell and close to the
    opposite :math:`xy`-plane of the box, i.e., shifted by
    :math:`\vec{t} = 0\cdot\vec{a}+0\cdot\vec{b}+1\cdot\vec{c}=\vec{c}`.

    Likewise, if the particle is close to more than one box walls, images along
    the diagonals are also generated::

                                    x            x
        +------------+                +------------+
        |            |   augment      |            |
        |            |   ------->     |            |
        |          o |              x |          o |
        +------------+                +------------+

    Parameters
    ----------
    coordinates : numpy.ndarray
      Input coordinate array of shape ``(n, 3)`` and dtype ``numpy.float32``
      used to generate duplicate images in the vicinity of the central cell. All
      coordinates must be within the primary unit cell.
    box : numpy.ndarray
      Box dimensions of shape ``(6,)`` and dtype ``numpy.float32``. The
      dimensions must be provided in the same format as returned
      by :attr:`MDAnalysis.coordinates.base.Timestep.dimensions`:
      ``[lx, ly, lz, alpha, beta, gamma]``
    r : float
      Thickness of cutoff region for duplicate image generation.

    Returns
    -------
    output : numpy.ndarray
      Coordinates of duplicate (augmented) particles (dtype ``numpy.float32``).
    indices : numpy.ndarray
      Original indices of the augmented coordinates (dtype ``numpy.int64``).
      Maps the indices of augmented particles to their original particle index
      such that ``indices[augmented_index] = original_index``.

    Note
    ----
    Output does not return coordinates from the initial array.
    To merge the particles with their respective images, the following operation
    is necessary when generating the images:

    .. code-block:: python

        images, mapping = augment_coordinates(coordinates, box, max_cutoff)
        all_coords = numpy.concatenate([coordinates, images])


    See Also
    --------
    :meth:`undo_augment`


    .. versionadded:: 0.19.0
    """
    cdef bint lo_x, hi_x, lo_y, hi_y, lo_z, hi_z
    cdef int i, j, N
    cdef float norm
    cdef float shiftX[3]
    cdef float shiftY[3]
    cdef float shiftZ[3]
    cdef float coord[3]
    cdef float end[3]
    cdef float other[3]
    cdef float dm[3][3]
    cdef float reciprocal[3][3]

    dm = triclinic_vectors(box)

    for i in range(3):
        shiftX[i] = dm[0][i]
        shiftY[i] = dm[1][i]
        shiftZ[i] = dm[2][i]
        end[i] = dm[0][i] + dm[1][i] + dm[2][i]
    # Calculate reciprocal vectors
    _cross(&dm[1][0], &dm[2][0], &reciprocal[0][0])
    _cross(&dm[2][0], &dm[0][0], &reciprocal[1][0])
    _cross(&dm[0][0], &dm[1][0], &reciprocal[2][0])
    # Normalize
    for i in range(3):
        norm = _norm(&reciprocal[i][0])
        for j in range(3):
            reciprocal[i][j] = reciprocal[i][j]/norm

    N = coordinates.shape[0]

    cdef vector[float] output
    cdef vector[int] indices

    for i in range(N):
        for j in range(3):
            coord[j] = coordinates[i, j]
            other[j] = end[j] - coordinates[i, j]
        # identify the condition
        lo_x = _dot(&coord[0], &reciprocal[0][0]) <= r
        hi_x = _dot(&other[0], &reciprocal[0][0]) <= r
        lo_y = _dot(&coord[0], &reciprocal[1][0]) <= r
        hi_y = _dot(&other[0], &reciprocal[1][0]) <= r
        lo_z = _dot(&coord[0], &reciprocal[2][0]) <= r
        hi_z = _dot(&other[0], &reciprocal[2][0]) <= r

        if lo_x:
            # if X, face piece
            for j in range(3):
                # add to output
                output.push_back(coord[j] + shiftX[j])
            # keep record of which index this augmented
            # position was created from
            indices.push_back(i)

            if lo_y:
                # if X&Y, edge piece
                for j in range(3):
                    output.push_back(coord[j] + shiftX[j] + shiftY[j])
                indices.push_back(i)

                if lo_z:
                    # if X&Y&Z, corner piece
                    for j in range(3):
                        output.push_back(coord[j] + shiftX[j] + shiftY[j] + shiftZ[j])
                    indices.push_back(i)

                elif hi_z:
                    for j in range(3):
                        output.push_back(coord[j] + shiftX[j] + shiftY[j] - shiftZ[j])
                    indices.push_back(i)

            elif hi_y:
                for j in range(3):
                    output.push_back(coord[j] + shiftX[j] - shiftY[j])
                indices.push_back(i)

                if lo_z:
                    for j in range(3):
                        output.push_back(coord[j] + shiftX[j] - shiftY[j] + shiftZ[j])
                    indices.push_back(i)

                elif hi_z:
                    for j in range(3):
                        output.push_back(coord[j] + shiftX[j] - shiftY[j] - shiftZ[j])
                    indices.push_back(i)

            if lo_z:
                for j in range(3):
                    output.push_back(coord[j] + shiftX[j] + shiftZ[j])
                indices.push_back(i)

            elif hi_z:
                for j in range(3):
                    output.push_back(coord[j] + shiftX[j] - shiftZ[j])
                indices.push_back(i)

        elif hi_x:
            for j in range(3):
                output.push_back(coord[j] - shiftX[j])
            indices.push_back(i)

            if lo_y:
                for j in range(3):
                    output.push_back(coord[j] - shiftX[j] + shiftY[j])
                indices.push_back(i)

                if lo_z:
                    for j in range(3):
                        output.push_back(coord[j] - shiftX[j] + shiftY[j] + shiftZ[j])
                    indices.push_back(i)

                elif hi_z:
                    for j in range(3):
                        output.push_back(coord[j] - shiftX[j] + shiftY[j] - shiftZ[j])
                    indices.push_back(i)

            elif hi_y:
                for j in range(3):
                    output.push_back(coord[j] - shiftX[j] - shiftY[j])
                indices.push_back(i)

                if lo_z:
                    for j in range(3):
                        output.push_back(coord[j] - shiftX[j] - shiftY[j] + shiftZ[j])
                    indices.push_back(i)

                elif hi_z:
                    for j in range(3):
                        output.push_back(coord[j] - shiftX[j] - shiftY[j] - shiftZ[j])
                    indices.push_back(i)

            if lo_z:
                for j in range(3):
                    output.push_back(coord[j] - shiftX[j] + shiftZ[j])
                indices.push_back(i)

            elif hi_z:
                for j in range(3):
                    output.push_back(coord[j] - shiftX[j] - shiftZ[j])
                indices.push_back(i)

        if lo_y:
            for j in range(3):
                output.push_back(coord[j] + shiftY[j])
            indices.push_back(i)

            if lo_z:
                for j in range(3):
                    output.push_back(coord[j] + shiftY[j] + shiftZ[j])
                indices.push_back(i)

            elif hi_z:
                for j in range(3):
                    output.push_back(coord[j] + shiftY[j] - shiftZ[j])
                indices.push_back(i)

        elif hi_y:
            for j in range(3):
                output.push_back(coord[j] - shiftY[j])
            indices.push_back(i)

            if lo_z:
                for j in range(3):
                    output.push_back(coord[j] - shiftY[j] + shiftZ[j])
                indices.push_back(i)

            elif hi_z:
                for j in range(3):
                    output.push_back(coord[j] - shiftY[j] - shiftZ[j])
                indices.push_back(i)

        if lo_z:
            for j in range(3):
                output.push_back(coord[j] + shiftZ[j])
            indices.push_back(i)

        elif hi_z:
            for j in range(3):
                output.push_back(coord[j] - shiftZ[j])
            indices.push_back(i)
    n = indices.size()
    return np.asarray(output, dtype=np.float32).reshape(n, 3), np.asarray(indices, dtype=np.intp)


@cython.boundscheck(False)
@cython.wraparound(False)
def undo_augment(cnp.intp_t[:] results, cnp.intp_t[:] translation, int nreal):
    """Translate augmented indices back to original indices.

    Parameters
    ----------
    results : numpy.ndarray
      Array of dtype ``numpy.int64`` containing coordinate indices, including
      "augmented" indices.
    translation : numpy.ndarray
      Index map of dtype ``numpy.int64`` linking the augmented indices to the
      original particle indices such that
      ``translation[augmented_index] = original_index``.
    nreal : int
      Number of real coordinates, i.e., indices in `results` equal or larger
      than this need to be mapped to their real counterpart.

    Returns
    -------
    results : numpy.ndarray
      Modified input `results` with all the augmented indices translated to
      their corresponding initial original indices.

    Note
    ----
    Modifies the results array in place.

    See Also
    --------
    :meth:`augment_coordinates`


    .. versionadded:: 0.19.0
    """
    cdef int N
    cdef ssize_t i
    N = results.shape[0]

    for i in range(N):
        if results[i] >= nreal:
            results[i] = translation[results[i] - nreal]
    return np.asarray(results, dtype=np.intp)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Base guesser classes --- :mod:`MDAnalysis.guesser.base`
================================================================

Derive context-specific guesser classes from the base class in this module.

Classes
-------

.. autoclass:: GuesserBase
   :members:
   :inherited-members:

.. autofunction:: get_guesser

"""
from .. import _GUESSERS, _TOPOLOGY_ATTRS
from ..core.topologyattrs import _Connection
import numpy as np
import logging
from typing import Dict
import copy

logger = logging.getLogger("MDAnalysis.guesser.base")


class _GuesserMeta(type):
    """Internal: guesser classes registration

    When classes which inherit from GuesserBase are *defined*
    this metaclass makes it known to MDAnalysis.  'context'
    attribute  are read:
     - `context` defines the context of the guesser class for example:
       forcefield specific context  as MartiniGuesser
       and file specific context as PDBGuesser.

    Eg::

      class FooGuesser(GuesserBase):
          format = 'foo'

    .. versionadded:: 2.8.0
    """

    def __init__(cls, name, bases, classdict):
        type.__init__(type, name, bases, classdict)

        _GUESSERS[classdict["context"].upper()] = cls


class GuesserBase(metaclass=_GuesserMeta):
    """Base class for context-specific guessers to inherit from

    Parameters
    ----------
    universe : Universe, optional
        Supply a Universe to the Guesser. This then becomes the source of atom
        attributes to be used in guessing processes. (this is relevant to how
        the universe's guess_TopologyAttrs API works.
        See :meth:`~MDAnalysis.core.universe.Universe.guess_TopologyAttrs`).
    **kwargs : dict, optional
        To pass additional data to the guesser that can be used with
              different methods.


    .. versionadded:: 2.8.0

    """

    context = "base"
    _guesser_methods: Dict = {}

    def __init__(self, universe=None, **kwargs):
        self._universe = universe
        self._kwargs = kwargs

    def update_kwargs(self, **kwargs):
        self._kwargs.update(kwargs)

    def copy(self):
        """Return a copy of this Guesser"""
        kwargs = copy.deepcopy(self._kwargs)
        new = self.__class__(universe=None, **kwargs)
        return new

    def is_guessable(self, attr_to_guess):
        """check if the passed atrribute can be guessed by the guesser class

        Parameters
        ----------
        guess: str
            Attribute to be guessed then added to the Universe

        Returns
        -------
        bool
        """
        if attr_to_guess.lower() in self._guesser_methods:
            return True

        return False

    def guess_attr(self, attr_to_guess, force_guess=False):
        """map the attribute to be guessed with the apporpiate guessing method

        Parameters
        ----------
        attr_to_guess: str
            an atrribute to be guessed then to be added to the universe
        force_guess: bool
            To indicate wether to only partialy guess the empty values of the
            attribute or to overwrite all existing values by guessed one

        Returns
        -------
        NDArray of guessed values

        """
        try:
            top_attr = _TOPOLOGY_ATTRS[attr_to_guess]
        except KeyError:
            raise KeyError(
                f"{attr_to_guess} is not a recognized MDAnalysis "
                "topology attribute"
            )
        # make attribute to guess plural
        attr_to_guess = top_attr.attrname

        try:
            guesser_method = self._guesser_methods[attr_to_guess]
        except KeyError:
            raise ValueError(
                f"{type(self).__name__} cannot guess this "
                f"attribute: {attr_to_guess}"
            )

        # Connection attributes should be just returned as they are always
        # appended to the Universe. ``force_guess`` handling should happen
        # at Universe level.
        if issubclass(top_attr, _Connection):
            return guesser_method()

        # check if the topology already has the attribute to partially guess it
        if hasattr(self._universe.atoms, attr_to_guess) and not force_guess:
            attr_values = np.array(
                getattr(self._universe.atoms, attr_to_guess, None)
            )

            empty_values = top_attr.are_values_missing(attr_values)

            if True in empty_values:
                # pass to the guesser_method boolean mask to only guess the
                # empty values
                attr_values[empty_values] = guesser_method(
                    indices_to_guess=empty_values
                )
                return attr_values

            else:
                logger.info(
                    f"There is no empty {attr_to_guess} values. Guesser did "
                    f"not guess any new values for {attr_to_guess} attribute"
                )
                return None
        else:
            return np.array(guesser_method())


def get_guesser(context, u=None, **kwargs):
    """get an appropiate guesser to the Universe and pass
       the Universe to the guesser

    Parameters
    ----------
    u: Universe
        to be passed to the guesser
    context: str or Guesser
    **kwargs : dict, optional
        Extra arguments are passed to the guesser.

    Returns
    -------
    Guesser class

    Raises
    ------
    * :exc:`KeyError` upon failing to return a guesser class

    .. versionadded:: 2.8.0

    """
    if isinstance(context, GuesserBase):
        context._universe = u
        context.update_kwargs(**kwargs)
        return context
    try:
        if issubclass(context, GuesserBase):
            return context(u, **kwargs)
    except TypeError:
        pass

    try:
        guesser = _GUESSERS[context.upper()](u, **kwargs)
    except KeyError:
        raise KeyError("Unidentified guesser type {0}".format(context))
    return guesser
